#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for costing package based on methods from:

    Process and Product Design Principles: Synthesis, Analysis, and
    Evaluation
    Seider, Seader, Lewin, Windagdo, 3rd Ed. John Wiley and Sons
    Chapter 22. Cost Accounting and Capital Cost Estimation
    22.2 Cost Indexes and Capital Investment
"""
import pytest

from pyomo.environ import (
    Block,
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    Param,
    units as pyunits,
    value,
    Var,
)
from pyomo.util.check_units import assert_units_consistent
from pyomo.common.config import ConfigValue

from idaes.core import FlowsheetBlock, UnitModelBlock, UnitModelCostingBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models.unit_models import (
    Compressor,
    CSTR,
    Flash,
    Heater,
    HeatExchanger,
    HeatExchangerNTU,
    PFR,
    PressureChanger,
    Pump,
    StoichiometricReactor,
    Turbine,
)
from idaes.models.unit_models.pressure_changer import ThermodynamicAssumption
from idaes.core.util.testing import (
    PhysicalParameterTestBlock,
    ReactionParameterTestBlock,
)
from idaes.models.properties import iapws95

from idaes.models.costing.SSLW import (
    SSLWCosting,
    SSLWCostingData,
    HXMaterial,
    HXTubeLength,
    HXType,
    VesselMaterial,
    TrayType,
    TrayMaterial,
    HeaterMaterial,
    HeaterSource,
    PumpType,
    PumpMaterial,
    PumpMotorType,
    CompressorDriveType,
    CompressorMaterial,
    CompressorType,
    FanMaterial,
    FanType,
    BlowerType,
    BlowerMaterial,
)
from idaes.models.properties.general_helmholtz import helmholtz_available


import logging
from io import StringIO
from pyomo.common.log import LoggingIntercept

# Some more information about this module
__author__ = "Andrew Lee"


solver = get_solver("ipopt_v2")


@pytest.fixture
def model():
    m = ConcreteModel()

    m.fs = FlowsheetBlock()

    m.fs.costing = SSLWCosting()

    # Add a placeholder to represent a unit model
    m.fs.unit = UnitModelBlock()

    return m


@pytest.mark.component
@pytest.mark.parametrize("material", HXMaterial)
@pytest.mark.parametrize("hxtype", HXType)
@pytest.mark.parametrize("tube_length", HXTubeLength)
def test_cost_heat_exchanger(model, material, hxtype, tube_length):
    model.fs.unit.area = Param(initialize=1000, units=pyunits.m**2)
    model.fs.unit.hot_side = Block()
    model.fs.unit.hot_side.properties_in = Block(model.fs.time)
    model.fs.unit.hot_side.properties_in[0].pressure = Param(
        initialize=2, units=pyunits.atm
    )

    model.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=model.fs.costing,
        costing_method=SSLWCostingData.cost_heat_exchanger,
        costing_method_arguments={
            "hx_type": hxtype,
            "material_type": material,
            "tube_length": tube_length,
        },
    )

    assert isinstance(model.fs.unit.costing.base_cost_per_unit, Var)
    assert isinstance(model.fs.unit.costing.capital_cost, Var)
    assert isinstance(model.fs.unit.costing.number_of_units, Var)
    assert isinstance(model.fs.unit.costing.pressure_factor, Var)
    assert isinstance(model.fs.unit.costing.material_factor, Var)

    assert isinstance(model.fs.unit.costing.capital_cost_constraint, Constraint)
    assert isinstance(model.fs.unit.costing.hx_material_eqn, Constraint)
    assert isinstance(model.fs.unit.costing.p_factor_eq, Constraint)

    assert degrees_of_freedom(model) == 0
    assert_units_consistent(model.fs.unit.costing)

    res = solver.solve(model)

    assert check_optimal_termination(res)

    # Test solution for one known case
    if (
        material == HXMaterial.StainlessSteelStainlessSteel
        and hxtype == HXType.Utube
        and tube_length == HXTubeLength.TwelveFoot
    ):
        assert pytest.approx(87704.6, 1e-5) == value(
            model.fs.unit.costing.base_cost_per_unit
        )
        assert pytest.approx(0.982982, 1e-5) == value(
            model.fs.unit.costing.pressure_factor
        )
        assert pytest.approx(4.08752, 1e-5) == value(
            model.fs.unit.costing.material_factor
        )

        assert pytest.approx(476063, 1e-5) == value(
            pyunits.convert(
                model.fs.unit.costing.capital_cost, to_units=pyunits.USD_2018
            )
        )


@pytest.mark.component
@pytest.mark.parametrize("material_type", VesselMaterial)
@pytest.mark.parametrize("weight_limit", [1, 2])
@pytest.mark.parametrize("aspect_ratio_range", [1, 2])
@pytest.mark.parametrize("include_pl", [True, False])
def test_cost_vessel(
    model, material_type, weight_limit, aspect_ratio_range, include_pl
):
    model.fs.unit.length = Param(initialize=0.00075, units=pyunits.m)
    model.fs.unit.diameter = Param(initialize=2, units=pyunits.m)

    model.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=model.fs.costing,
        costing_method=SSLWCostingData.cost_vessel,
        costing_method_arguments={
            "vertical": True,
            "material_type": material_type,
            "weight_limit": weight_limit,
            "aspect_ratio_range": aspect_ratio_range,
            "include_platforms_ladders": include_pl,
        },
    )

    assert isinstance(model.fs.unit.costing.shell_thickness, Param)
    assert isinstance(model.fs.unit.costing.material_factor, Param)
    assert isinstance(model.fs.unit.costing.material_density, Param)

    assert isinstance(model.fs.unit.costing.base_cost_per_unit, Var)
    assert isinstance(model.fs.unit.costing.capital_cost, Var)
    assert isinstance(model.fs.unit.costing.weight, Var)

    assert isinstance(model.fs.unit.costing.capital_cost_constraint, Constraint)
    assert isinstance(model.fs.unit.costing.base_cost_constraint, Constraint)
    assert isinstance(model.fs.unit.costing.weight_eq, Constraint)

    # Platforms and ladders
    if include_pl:
        assert isinstance(model.fs.unit.costing.base_cost_platforms_ladders, Var)

        assert isinstance(model.fs.unit.costing.cost_platforms_ladders_eq, Constraint)

    assert degrees_of_freedom(model) == 0
    assert_units_consistent(model.fs.unit.costing)

    res = solver.solve(model)

    assert check_optimal_termination(res)

    # Test solution for one known case
    if (
        material_type == VesselMaterial.CarbonSteel
        and weight_limit == 1
        and aspect_ratio_range == 1
        and include_pl
    ):
        assert pytest.approx(35958, 1e-5) == value(
            pyunits.convert(
                model.fs.unit.costing.capital_cost, to_units=pyunits.USD_2018
            )
        )


@pytest.mark.component
@pytest.mark.parametrize("tray_material", TrayMaterial)
@pytest.mark.parametrize("tray_type", TrayType)
def test_cost_vessel_trays(model, tray_material, tray_type):
    model.fs.unit.length = Param(initialize=0.00075, units=pyunits.m)
    model.fs.unit.diameter = Param(initialize=2, units=pyunits.m)

    model.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=model.fs.costing,
        costing_method=SSLWCostingData.cost_vessel,
        costing_method_arguments={
            "vertical": True,
            "number_of_trays": 10,
            "tray_material": tray_material,
            "tray_type": tray_type,
        },
    )

    assert isinstance(model.fs.unit.costing.shell_thickness, Param)
    assert isinstance(model.fs.unit.costing.material_factor, Param)
    assert isinstance(model.fs.unit.costing.material_density, Param)

    assert isinstance(model.fs.unit.costing.base_cost_per_unit, Var)
    assert isinstance(model.fs.unit.costing.capital_cost, Var)
    assert isinstance(model.fs.unit.costing.weight, Var)

    assert isinstance(model.fs.unit.costing.capital_cost_constraint, Constraint)
    assert isinstance(model.fs.unit.costing.base_cost_constraint, Constraint)
    assert isinstance(model.fs.unit.costing.weight_eq, Constraint)

    assert isinstance(model.fs.unit.costing.base_cost_platforms_ladders, Var)

    assert isinstance(model.fs.unit.costing.cost_platforms_ladders_eq, Constraint)

    assert isinstance(model.fs.unit.costing.tray_type_factor, Param)

    assert isinstance(model.fs.unit.costing.base_cost_trays, Var)
    assert isinstance(model.fs.unit.costing.tray_material_factor, Var)
    assert isinstance(model.fs.unit.costing.number_trays_factor, Var)
    assert isinstance(model.fs.unit.costing.base_cost_per_tray, Var)

    assert isinstance(model.fs.unit.costing.tray_material_factor_eq, Constraint)
    assert isinstance(model.fs.unit.costing.num_tray_factor_constraint, Constraint)
    assert isinstance(model.fs.unit.costing.single_tray_cost_constraint, Constraint)
    assert isinstance(model.fs.unit.costing.tray_costing_constraint, Constraint)

    assert degrees_of_freedom(model) == 0
    assert_units_consistent(model.fs.unit.costing)

    res = solver.solve(model)

    assert check_optimal_termination(res)


@pytest.mark.component
@pytest.mark.parametrize("material_type", VesselMaterial)
def test_cost_vessel_horizontal(model, material_type):
    model.fs.unit.length = Param(initialize=0.00075, units=pyunits.m)
    model.fs.unit.diameter = Param(initialize=2, units=pyunits.m)

    model.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=model.fs.costing,
        costing_method=SSLWCostingData.cost_vessel,
        costing_method_arguments={
            "vertical": False,
            "material_type": material_type,
            "weight_limit": 1,
            "include_platforms_ladders": False,
        },
    )

    assert isinstance(model.fs.unit.costing.shell_thickness, Param)
    assert isinstance(model.fs.unit.costing.material_factor, Param)
    assert isinstance(model.fs.unit.costing.material_density, Param)

    assert isinstance(model.fs.unit.costing.base_cost_per_unit, Var)
    assert isinstance(model.fs.unit.costing.capital_cost, Var)
    assert isinstance(model.fs.unit.costing.weight, Var)

    assert isinstance(model.fs.unit.costing.capital_cost_constraint, Constraint)
    assert isinstance(model.fs.unit.costing.base_cost_constraint, Constraint)
    assert isinstance(model.fs.unit.costing.weight_eq, Constraint)

    assert degrees_of_freedom(model) == 0
    assert_units_consistent(model.fs.unit.costing)

    res = solver.solve(model)

    assert check_optimal_termination(res)


@pytest.mark.component
@pytest.mark.parametrize("material_type", HeaterMaterial)
@pytest.mark.parametrize("heat_source", HeaterSource)
def test_cost_fired_heater(model, material_type, heat_source):
    model.fs.unit.heat_duty = Param([0], initialize=1000, units=pyunits.kJ / pyunits.s)
    model.fs.unit.control_volume = Block()
    model.fs.unit.control_volume.properties_in = Block(model.fs.time)
    model.fs.unit.control_volume.properties_in[0].pressure = Param(
        initialize=2, units=pyunits.atm
    )

    model.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=model.fs.costing,
        costing_method=SSLWCostingData.cost_fired_heater,
        costing_method_arguments={
            "material_type": material_type,
            "heat_source": heat_source,
        },
    )

    assert isinstance(model.fs.unit.costing.pressure_factor, Var)
    assert isinstance(model.fs.unit.costing.base_cost_per_unit, Var)
    assert isinstance(model.fs.unit.costing.capital_cost, Var)

    assert isinstance(model.fs.unit.costing.capital_cost_constraint, Constraint)
    assert isinstance(model.fs.unit.costing.base_cost_per_unit_eq, Constraint)
    assert isinstance(model.fs.unit.costing.pressure_factor_eq, Constraint)

    assert degrees_of_freedom(model) == 0
    assert_units_consistent(model.fs.unit.costing)

    res = solver.solve(model)

    assert check_optimal_termination(res)


@pytest.mark.component
def test_cost_turbine(model):
    model.fs.unit.config.declare("compressor", ConfigValue(default=False))
    model.fs.unit.work_mechanical = Param(
        [0], initialize=-1224638, units=pyunits.J / pyunits.s
    )

    model.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=model.fs.costing,
        costing_method=SSLWCostingData.cost_turbine,
    )

    assert isinstance(model.fs.unit.costing.capital_cost, Var)

    assert isinstance(model.fs.unit.costing.capital_cost_constraint, Constraint)

    assert degrees_of_freedom(model) == 0
    assert_units_consistent(model.fs.unit.costing)

    res = solver.solve(model)

    assert check_optimal_termination(res)

    assert pytest.approx(213199, 1e-5) == value(model.fs.unit.costing.capital_cost)


@pytest.mark.component
@pytest.mark.parametrize(
    "material_type",
    [
        PumpMaterial.CastIron,
        PumpMaterial.DuctileIron,
        PumpMaterial.CastSteel,
        PumpMaterial.Bronze,
        PumpMaterial.StainlessSteel,
        PumpMaterial.HastelloyC,
        PumpMaterial.Monel,
        PumpMaterial.Nickel,
        PumpMaterial.Titanium,
    ],
)
@pytest.mark.parametrize("pump_type_factor", [1.1, 1.2, 1.3, 1.4, 2.1, 2.2])
@pytest.mark.parametrize("motor_type", PumpMotorType)
def test_cost_pump_centrifugal(model, material_type, pump_type_factor, motor_type):
    model.fs.unit.config.declare("compressor", ConfigValue(default=True))
    model.fs.unit.work_mechanical = Param(
        [0], initialize=10143.79, units=pyunits.J / pyunits.s
    )
    model.fs.unit.deltaP = Param([0], initialize=50000, units=pyunits.Pa)
    model.fs.unit.control_volume = Block()
    model.fs.unit.control_volume.properties_in = Block(model.fs.time)
    model.fs.unit.control_volume.properties_in[0].dens_mass = Param(
        initialize=986.64, units=pyunits.kg / pyunits.m**3
    )
    model.fs.unit.control_volume.properties_in[0].flow_vol = Param(
        initialize=0.182592, units=pyunits.m**3 / pyunits.s
    )

    model.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=model.fs.costing,
        costing_method=SSLWCostingData.cost_pump,
        costing_method_arguments={
            "pump_type": PumpType.Centrifugal,
            "material_type": material_type,
            "pump_type_factor": pump_type_factor,
            "motor_type": motor_type,
        },
    )

    assert isinstance(model.fs.unit.costing.capital_cost, Var)
    assert isinstance(model.fs.unit.costing.number_of_units, Var)
    assert isinstance(model.fs.unit.costing.pump_head, Var)
    assert isinstance(model.fs.unit.costing.size_factor, Var)
    assert isinstance(model.fs.unit.costing.base_pump_cost_per_unit, Var)
    assert isinstance(model.fs.unit.costing.pump_capital_cost, Var)
    assert isinstance(model.fs.unit.costing.base_motor_cost_per_unit, Var)
    assert isinstance(model.fs.unit.costing.motor_capital_cost, Var)

    assert isinstance(model.fs.unit.costing.material_factor, Param)
    assert isinstance(model.fs.unit.costing.FT, Param)
    assert isinstance(model.fs.unit.costing.motor_FT, Param)

    assert isinstance(model.fs.unit.costing.capital_cost_constraint, Constraint)
    assert isinstance(model.fs.unit.costing.base_pump_cost_per_unit_eq, Constraint)
    assert isinstance(model.fs.unit.costing.base_motor_cost_eq, Constraint)
    assert isinstance(model.fs.unit.costing.pump_capital_cost_eq, Constraint)
    assert isinstance(model.fs.unit.costing.motor_capital_cost_eq, Constraint)

    assert degrees_of_freedom(model) == 0
    assert_units_consistent(model.fs.unit.costing)

    res = solver.solve(model)

    assert check_optimal_termination(res)

    # Test solution for one known case
    if (
        material_type == PumpMaterial.Nickel
        and pump_type_factor == 1.4
        and motor_type == PumpMotorType.Enclosed
    ):
        assert pytest.approx(62868.3, 1e-5) == value(
            pyunits.convert(
                model.fs.unit.costing.capital_cost, to_units=pyunits.USD_2018
            )
        )


@pytest.mark.component
@pytest.mark.parametrize(
    "material_type",
    [
        PumpMaterial.CastIron,
        PumpMaterial.DuctileIron,
        PumpMaterial.CastSteel,
        PumpMaterial.Bronze,
        PumpMaterial.StainlessSteel,
        PumpMaterial.HastelloyC,
        PumpMaterial.Monel,
        PumpMaterial.Nickel,
        PumpMaterial.Titanium,
    ],
)
@pytest.mark.parametrize("motor_type", PumpMotorType)
def test_cost_pump_ExternalGear(model, material_type, motor_type):
    model.fs.unit.config.declare("compressor", ConfigValue(default=True))
    model.fs.unit.work_mechanical = Param(
        [0], initialize=10143.79, units=pyunits.J / pyunits.s
    )
    model.fs.unit.deltaP = Param([0], initialize=50000, units=pyunits.Pa)
    model.fs.unit.control_volume = Block()
    model.fs.unit.control_volume.properties_in = Block(model.fs.time)
    model.fs.unit.control_volume.properties_in[0].dens_mass = Param(
        initialize=986.64, units=pyunits.kg / pyunits.m**3
    )
    model.fs.unit.control_volume.properties_in[0].flow_vol = Param(
        initialize=0.182592, units=pyunits.m**3 / pyunits.s
    )

    model.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=model.fs.costing,
        costing_method=SSLWCostingData.cost_pump,
        costing_method_arguments={
            "pump_type": PumpType.ExternalGear,
            "material_type": material_type,
            "motor_type": motor_type,
        },
    )

    assert isinstance(model.fs.unit.costing.capital_cost, Var)
    assert isinstance(model.fs.unit.costing.number_of_units, Var)
    assert isinstance(model.fs.unit.costing.pump_head, Var)
    assert isinstance(model.fs.unit.costing.size_factor, Var)
    assert isinstance(model.fs.unit.costing.base_pump_cost_per_unit, Var)
    assert isinstance(model.fs.unit.costing.pump_capital_cost, Var)
    assert isinstance(model.fs.unit.costing.base_motor_cost_per_unit, Var)
    assert isinstance(model.fs.unit.costing.motor_capital_cost, Var)

    assert isinstance(model.fs.unit.costing.material_factor, Param)
    assert isinstance(model.fs.unit.costing.FT, Param)
    assert isinstance(model.fs.unit.costing.motor_FT, Param)

    assert isinstance(model.fs.unit.costing.capital_cost_constraint, Constraint)
    assert isinstance(model.fs.unit.costing.base_pump_cost_per_unit_eq, Constraint)
    assert isinstance(model.fs.unit.costing.base_motor_cost_eq, Constraint)
    assert isinstance(model.fs.unit.costing.pump_capital_cost_eq, Constraint)
    assert isinstance(model.fs.unit.costing.motor_capital_cost_eq, Constraint)

    assert degrees_of_freedom(model) == 0
    assert_units_consistent(model.fs.unit.costing)

    res = solver.solve(model)

    assert check_optimal_termination(res)


@pytest.mark.component
@pytest.mark.parametrize(
    "material_type",
    [
        PumpMaterial.DuctileIron,
        PumpMaterial.StainlessSteel,
        PumpMaterial.NiAlBronze,
        PumpMaterial.CarbonSteel,
    ],
)
@pytest.mark.parametrize("motor_type", PumpMotorType)
def test_cost_pump_reciprocating(model, material_type, motor_type):
    model.fs.unit.config.declare("compressor", ConfigValue(default=True))
    model.fs.unit.work_mechanical = Param(
        [0], initialize=10143.79, units=pyunits.J / pyunits.s
    )
    model.fs.unit.deltaP = Param([0], initialize=50000, units=pyunits.Pa)
    model.fs.unit.control_volume = Block()
    model.fs.unit.control_volume.properties_in = Block(model.fs.time)
    model.fs.unit.control_volume.properties_in[0].dens_mass = Param(
        initialize=986.64, units=pyunits.kg / pyunits.m**3
    )
    model.fs.unit.control_volume.properties_in[0].flow_vol = Param(
        initialize=0.182592, units=pyunits.m**3 / pyunits.s
    )

    model.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=model.fs.costing,
        costing_method=SSLWCostingData.cost_pump,
        costing_method_arguments={
            "pump_type": PumpType.Reciprocating,
            "material_type": material_type,
            "motor_type": motor_type,
        },
    )

    assert isinstance(model.fs.unit.costing.capital_cost, Var)
    assert isinstance(model.fs.unit.costing.number_of_units, Var)
    assert isinstance(model.fs.unit.costing.pump_head, Var)
    assert isinstance(model.fs.unit.costing.size_factor, Var)
    assert isinstance(model.fs.unit.costing.base_pump_cost_per_unit, Var)
    assert isinstance(model.fs.unit.costing.pump_capital_cost, Var)
    assert isinstance(model.fs.unit.costing.base_motor_cost_per_unit, Var)
    assert isinstance(model.fs.unit.costing.motor_capital_cost, Var)

    assert isinstance(model.fs.unit.costing.material_factor, Param)
    assert isinstance(model.fs.unit.costing.FT, Param)
    assert isinstance(model.fs.unit.costing.motor_FT, Param)

    assert isinstance(model.fs.unit.costing.capital_cost_constraint, Constraint)
    assert isinstance(model.fs.unit.costing.base_pump_cost_per_unit_eq, Constraint)
    assert isinstance(model.fs.unit.costing.base_motor_cost_eq, Constraint)
    assert isinstance(model.fs.unit.costing.pump_capital_cost_eq, Constraint)
    assert isinstance(model.fs.unit.costing.motor_capital_cost_eq, Constraint)

    assert degrees_of_freedom(model) == 0
    assert_units_consistent(model.fs.unit.costing)

    res = solver.solve(model)

    assert check_optimal_termination(res)


@pytest.mark.component
@pytest.mark.parametrize("compressor_type", CompressorType)
@pytest.mark.parametrize("drive_type", CompressorDriveType)
@pytest.mark.parametrize("material_type", CompressorMaterial)
def test_cost_compressor(model, compressor_type, drive_type, material_type):
    model.fs.unit.config.declare("compressor", ConfigValue(default=True))
    model.fs.unit.config.declare(
        "thermodynamic_assumption",
        ConfigValue(default=ThermodynamicAssumption.isothermal),
    )
    model.fs.unit.config.thermodynamic_assumption = ThermodynamicAssumption.isentropic
    model.fs.unit.work_mechanical = Param(
        [0], initialize=101410.4, units=pyunits.J / pyunits.s
    )

    model.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=model.fs.costing,
        costing_method=SSLWCostingData.cost_compressor,
        costing_method_arguments={
            "compressor_type": compressor_type,
            "drive_type": drive_type,
            "material_type": material_type,
        },
    )

    assert isinstance(model.fs.unit.costing.capital_cost, Var)

    assert isinstance(model.fs.unit.costing.capital_cost_constraint, Constraint)

    assert degrees_of_freedom(model) == 0
    assert_units_consistent(model.fs.unit.costing)

    res = solver.solve(model)

    assert check_optimal_termination(res)

    if (
        compressor_type == CompressorType.Centrifugal
        and drive_type == CompressorDriveType.ElectricMotor
        and material_type == CompressorMaterial.StainlessSteel
    ):
        assert pytest.approx(300695, 1e-5) == value(
            pyunits.convert(
                model.fs.unit.costing.capital_cost, to_units=pyunits.USD_2018
            )
        )


@pytest.mark.component
@pytest.mark.parametrize("compressor_type", CompressorType)
@pytest.mark.parametrize("drive_type", CompressorDriveType)
@pytest.mark.parametrize("material_type", CompressorMaterial)
def test_not_compressor(model, compressor_type, drive_type, material_type):
    # Test exception for non-supported compressor flags
    model.fs.unit.config.declare("compressor", ConfigValue(default=False))
    model.fs.unit.config.declare(
        "thermodynamic_assumption",
        ConfigValue(default=ThermodynamicAssumption.isentropic),
    )
    model.fs.unit.work_mechanical = Param(
        [0], initialize=101410.4, units=pyunits.J / pyunits.s
    )

    expected_string = (
        "cost_compressor method is only appropriate for "
        "pressure changers with the compressor argument "
        "equal to True."
    )

    stream = StringIO()
    with LoggingIntercept(stream, "idaes", logging.WARNING):
        model.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=model.fs.costing,
            costing_method=SSLWCostingData.cost_compressor,
            costing_method_arguments={
                "compressor_type": compressor_type,
                "drive_type": drive_type,
                "material_type": material_type,
            },
        )

    assert expected_string in str(stream.getvalue())


@pytest.mark.component
@pytest.mark.parametrize("fan_type", FanType)
@pytest.mark.parametrize("material_type", FanMaterial)
def test_cost_fan(model, fan_type, material_type):
    model.fs.unit.config.declare("compressor", ConfigValue(default=True))
    model.fs.unit.control_volume = Block()
    model.fs.unit.control_volume.properties_in = Block(model.fs.time)
    model.fs.unit.control_volume.properties_in[0].flow_vol = Param(
        initialize=0.182592, units=pyunits.m**3 / pyunits.s
    )

    model.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=model.fs.costing,
        costing_method=SSLWCostingData.cost_fan,
        costing_method_arguments={"fan_type": fan_type, "material_type": material_type},
    )

    assert isinstance(model.fs.unit.costing.capital_cost, Var)
    assert isinstance(model.fs.unit.costing.number_of_units, Var)
    assert isinstance(model.fs.unit.costing.base_cost_per_unit, Var)

    assert isinstance(model.fs.unit.costing.head_factor, Param)
    assert isinstance(model.fs.unit.costing.material_factor, Param)

    assert isinstance(model.fs.unit.costing.capital_cost_constraint, Constraint)
    assert isinstance(model.fs.unit.costing.base_cost_per_unit_eq, Constraint)

    assert degrees_of_freedom(model) == 0
    assert_units_consistent(model.fs.unit.costing)

    res = solver.solve(model)

    assert check_optimal_termination(res)

    # TODO: Test case for solution checking


@pytest.mark.component
@pytest.mark.parametrize("blower_type", BlowerType)
@pytest.mark.parametrize("material_type", BlowerMaterial)
def test_cost_blower(model, blower_type, material_type):
    model.fs.unit.config.declare("compressor", ConfigValue(default=True))
    model.fs.unit.work_mechanical = Param(
        [0], initialize=101410.4, units=pyunits.J / pyunits.s
    )

    model.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=model.fs.costing,
        costing_method=SSLWCostingData.cost_blower,
        costing_method_arguments={
            "blower_type": blower_type,
            "material_type": material_type,
        },
    )

    assert isinstance(model.fs.unit.costing.capital_cost, Var)
    assert isinstance(model.fs.unit.costing.number_of_units, Var)
    assert isinstance(model.fs.unit.costing.base_cost_per_unit, Var)

    assert isinstance(model.fs.unit.costing.material_factor, Param)

    assert isinstance(model.fs.unit.costing.capital_cost_constraint, Constraint)
    assert isinstance(model.fs.unit.costing.base_cost_per_unit_eq, Constraint)

    assert degrees_of_freedom(model) == 0
    assert_units_consistent(model.fs.unit.costing)

    res = solver.solve(model)

    assert check_optimal_termination(res)

    # TODO: Test case for solution checking


@pytest.mark.integration
class TestMapping:
    @pytest.fixture
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock()

        m.fs.pparams = PhysicalParameterTestBlock()
        m.fs.rparams = ReactionParameterTestBlock(property_package=m.fs.pparams)

        m.fs.costing = SSLWCosting()

        return m

    def test_compressor(self, model):
        # Add examples of supported unit models and add costing
        model.fs.C101 = Compressor(property_package=model.fs.pparams)
        model.fs.C101.costing = UnitModelCostingBlock(
            flowsheet_costing_block=model.fs.costing
        )
        assert model.fs.C101.costing.drive_factor.value == 1
        assert model.fs.C101.costing.material_factor.value == 2.5

    def test_cstr(self, model):
        model.fs.R102 = CSTR(
            property_package=model.fs.pparams, reaction_package=model.fs.rparams
        )
        # Add length and diameter to CSTR
        model.fs.R102.length = 1 * pyunits.m
        model.fs.R102.diameter = 1 * pyunits.m
        model.fs.R102.costing = UnitModelCostingBlock(
            flowsheet_costing_block=model.fs.costing
        )
        assert model.fs.R102.costing.material_factor.value == 1
        assert model.fs.R102.costing.material_density.value == 0.284

    def test_flash(self, model):
        model.fs.F103 = Flash(property_package=model.fs.pparams)
        # Add length and diameter to Flash
        model.fs.F103.length = 1 * pyunits.m
        model.fs.F103.diameter = 1 * pyunits.m
        model.fs.F103.costing = UnitModelCostingBlock(
            flowsheet_costing_block=model.fs.costing
        )
        assert model.fs.F103.costing.material_factor.value == 1
        assert model.fs.F103.costing.material_density.value == 0.284

    def test_heater(self, model):
        # Add examples of supported unit models and add costing
        model.fs.unit = Heater(property_package=model.fs.pparams)
        model.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=model.fs.costing
        )
        assert model.fs.unit.costing.pressure_factor.value == 1.1
        assert model.fs.unit.costing.material_factor.value == 1

    def test_hx0D(self, model):
        # Add examples of supported unit models and add costing
        model.fs.unit = HeatExchanger(
            hot_side={"property_package": model.fs.pparams},
            cold_side={"property_package": model.fs.pparams},
        )
        model.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=model.fs.costing
        )
        assert model.fs.unit.costing.length_factor.value == 1.12

    # TODO : Test for HX1D once supported

    @pytest.mark.skipif(
        not helmholtz_available(), reason="General Helmholtz not available"
    )
    def test_hx_ntu(self):
        # Need a different property package here
        m = ConcreteModel()

        m.fs = FlowsheetBlock()

        m.fs.pparams = iapws95.Iapws95ParameterBlock()

        m.fs.costing = SSLWCosting()
        # Add examples of supported unit models and add costing
        m.fs.unit = HeatExchangerNTU(
            hot_side={"property_package": m.fs.pparams},
            cold_side={"property_package": m.fs.pparams},
        )
        m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        assert m.fs.unit.costing.length_factor.value == 1.12

    def test_pfr(self, model):
        model.fs.unit = PFR(
            property_package=model.fs.pparams, reaction_package=model.fs.rparams
        )
        # Add length and diameter to PFR
        model.fs.unit.length = 1 * pyunits.m
        model.fs.unit.diameter = 1 * pyunits.m
        model.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=model.fs.costing
        )
        assert model.fs.unit.costing.material_factor.value == 1
        assert model.fs.unit.costing.material_density.value == 0.284

    def test_pressure_changer(self, model):
        # Add examples of supported unit models and add costing
        model.fs.unit = PressureChanger(
            property_package=model.fs.pparams,
            thermodynamic_assumption=ThermodynamicAssumption.isentropic,
        )
        model.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=model.fs.costing
        )
        assert model.fs.unit.costing.drive_factor.value == 1
        assert model.fs.unit.costing.material_factor.value == 2.5

    def test_pump_compressor(self, model):
        # Test exception for non-supported compressor flags
        model.fs.unit = Compressor(
            property_package=model.fs.pparams,
            thermodynamic_assumption=ThermodynamicAssumption.pump,
        )

        expected_string = (
            "fs.unit - pressure changers with the pump "
            "assumption should use the cost_pump method."
        )

        stream = StringIO()
        with LoggingIntercept(stream, "idaes", logging.WARNING):
            model.fs.unit.costing = UnitModelCostingBlock(
                flowsheet_costing_block=model.fs.costing
            )

        assert expected_string in str(stream.getvalue())

    def test_isothermal_compressor(self, model):
        # Test exception for non-supported compressor flags
        model.fs.unit = Compressor(
            property_package=model.fs.pparams,
            thermodynamic_assumption=ThermodynamicAssumption.isothermal,
        )

        expected_string = (
            "fs.unit - pressure changers without isentropic "
            "assumption are too simple to be costed."
        )

        stream = StringIO()
        with LoggingIntercept(stream, "idaes", logging.WARNING):
            model.fs.unit.costing = UnitModelCostingBlock(
                flowsheet_costing_block=model.fs.costing
            )

        assert expected_string in str(stream.getvalue())

    def test_adiabatic_compressor(self, model):
        # Test exception for non-supported compressor flags
        model.fs.unit = Compressor(
            property_package=model.fs.pparams,
            thermodynamic_assumption=ThermodynamicAssumption.adiabatic,
        )

        expected_string = (
            "fs.unit - pressure changers without isentropic "
            "assumption are too simple to be costed."
        )

        stream = StringIO()
        with LoggingIntercept(stream, "idaes", logging.WARNING):
            model.fs.unit.costing = UnitModelCostingBlock(
                flowsheet_costing_block=model.fs.costing
            )

        assert expected_string in str(stream.getvalue())

    @pytest.mark.skipif(
        not helmholtz_available(), reason="General Helmholtz not available"
    )
    def test_pump(self):
        # Need a different property package here
        m = ConcreteModel()

        m.fs = FlowsheetBlock()

        m.fs.pparams = iapws95.Iapws95ParameterBlock()

        m.fs.costing = SSLWCosting()

        # Add examples of supported unit models and add costing
        m.fs.unit = Pump(property_package=m.fs.pparams)
        m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        assert hasattr(m.fs.unit.costing, "pump_head")

    def test_rstoich(self, model):
        model.fs.unit = StoichiometricReactor(
            property_package=model.fs.pparams, reaction_package=model.fs.rparams
        )
        # Add length and diameter to reactor
        model.fs.unit.length = 1 * pyunits.m
        model.fs.unit.diameter = 1 * pyunits.m
        model.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=model.fs.costing
        )
        assert model.fs.unit.costing.material_factor.value == 1
        assert model.fs.unit.costing.material_density.value == 0.284

    def test_turbine(self, model):
        # Add examples of supported unit models and add costing
        model.fs.unit = Turbine(property_package=model.fs.pparams)
        model.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=model.fs.costing
        )
        assert hasattr(model.fs.unit.costing, "capital_cost")
        assert not hasattr(model.fs.unit.costing, "material_factor")



# -*- coding: utf-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
""" Block-like object meant for models where almost everything is indexed
by time.
"""

from collections import OrderedDict

import idaes.logger as idaeslog
from idaes.apps.caprese.util import initialize_by_element_in_range
from idaes.apps.caprese.common.config import (
    InputOption,
)
from idaes.apps.caprese.common.config import VariableCategory as VC
from idaes.apps.caprese.categorize import (
    categorize_dae_variables,
    CATEGORY_TYPE_MAP,
)
from idaes.apps.caprese.nmpc_var import (
    NmpcVar,
    _NmpcVector,
    DiffVar,
    DerivVar,
    AlgVar,
    InputVar,
    FixedVar,
    MeasuredVar,
)
from idaes.core.util.model_statistics import degrees_of_freedom

from pyomo.environ import (
    Block,
    Reference,
    TransformationFactory,
    Var,
    Set,
    ComponentUID,
    Suffix,
)
from pyomo.core.base.initializer import Initializer
from pyomo.core.base.block import BlockData, SubclassOf
from pyomo.core.base.indexed_component import UnindexedComponent_set
from pyomo.common.collections import ComponentMap
from pyomo.common.config import ConfigDict, ConfigValue
from pyomo.core.base.range import remainder
from pyomo.dae.flatten import flatten_dae_components


class _DynamicBlockData(BlockData):
    """This class adds methods and data structures that are useful
    for working with dynamic models. These include methods for
    initialization and references to time-indexed variables.
    """

    # TODO: This class should probably give the option to clone
    # the user's model.

    logger = idaeslog.getLogger("nmpc")

    CONFIG = ConfigDict()

    CONFIG.declare(
        "tee",
        ConfigValue(
            default=True,
            domain=bool,
            doc="tee option for embedded solver calls",
        ),
    )

    CONFIG.declare(
        "outlvl",
        ConfigValue(
            default=idaeslog.INFO,
            doc="Output level for IDAES logger",
        ),
    )

    def _construct(self):
        """Generates time-indexed references and categorizes them."""
        model = self.mod
        time = self.time
        try:
            inputs = self._inputs
        except AttributeError:
            inputs = self._inputs = None
        try:
            measurements = self._measurements
        except AttributeError:
            measurements = self._measurements = None

        # TODO: Give the user the option to provide their own
        # category_dict (they know the structure of their model
        # better than I do...)
        if self._category_dict is not None:
            category_dict = self.category_dict = self._category_dict
            if VC.INPUT not in category_dict and inputs is not None:
                self.category_dict[VC.INPUT] = inputs
            if VC.MEASUREMENT not in category_dict and measurements is not None:
                self.category_dict[VC.MEASUREMENT] = measurements
            self.dae_vars = []
            for categ, varlist in category_dict.items():
                if categ is not VC.MEASUREMENT:
                    # Assume that measurements are duplicates
                    self.dae_vars.extend(varlist)

        else:
            scalar_vars, dae_vars = flatten_dae_components(
                model,
                time,
                ctype=Var,
            )
            self.scalar_vars = scalar_vars
            self.dae_vars = dae_vars
            category_dict = categorize_dae_variables(
                dae_vars,
                time,
                inputs,
                measurements=measurements,
            )
            self.category_dict = category_dict

        keys = list(category_dict)
        for categ in keys:
            if not category_dict[categ]:
                # Empty categories cause problems for us down the road
                # due empty (unknown dimension) slices.
                category_dict.pop(categ)

        self._add_category_blocks()
        self._add_category_references()

        self.categories = set(category_dict)

        # Now that we don't rely on knowing what the categories
        # will be, we do not need these attributes. They are, however,
        # used in the tests, so for now, we add them if possible.
        if VC.DIFFERENTIAL in category_dict:
            self.differential_vars = category_dict[VC.DIFFERENTIAL]
        if VC.ALGEBRAIC in category_dict:
            self.algebraic_vars = category_dict[VC.ALGEBRAIC]
        if VC.DERIVATIVE in category_dict:
            self.derivative_vars = category_dict[VC.DERIVATIVE]
        if VC.INPUT in category_dict:
            self.input_vars = category_dict[VC.INPUT]
        if VC.FIXED in category_dict:
            self.fixed_vars = category_dict[VC.FIXED]
        if VC.MEASUREMENT in category_dict:
            self.measurement_vars = category_dict.pop(VC.MEASUREMENT)

        # The categories in category_dict now form a partition of the
        # time-indexed variables. This is necessary to have a well-defined
        # vardata map, which maps each vardata to a unique component indexed
        # only by time.

        # Maps each vardata (of a time-indexed var) to the NmpcVar
        # that contains it.
        self.vardata_map = ComponentMap(
            (var[t], var)
            for var in self.component_objects(SubclassOf(NmpcVar))
            # for varlist in category_dict.values()
            # for var in varlist
            for t in time
            if var.ctype is not MeasuredVar
        )
        # NOTE: looking up var[t] instead of iterating over values()
        # appears to be ~ 5x faster

        # These should be overridden by a call to `set_sample_time`
        # The defaults assume that the entire model is one sample.
        self.sample_points = [time.first(), time.last()]
        self.sample_point_indices = [1, len(time)]

    _var_name = "var"
    _block_suffix = "_BLOCK"
    _set_suffix = "_SET"

    @classmethod
    def get_category_block_name(cls, categ):
        """Gets block name from name of enum entry"""
        return categ.name + cls._block_suffix

    @classmethod
    def get_category_set_name(cls, categ):
        """Gets set name from name of enum entry"""
        return categ.name + cls._set_suffix

    def _add_category_blocks(self):
        """Adds an indexed block for each category of variable and
        attach a reference to each variable to one of the BlockDatas.
        """
        category_dict = self.category_dict
        var_name = self._var_name
        for categ, varlist in category_dict.items():
            ctype = CATEGORY_TYPE_MAP.get(categ, NmpcVar)
            # These names are e.g. 'DIFFERENTIAL_BLOCK', 'DIFFERENTIAL_SET'
            # They serve as a way to access all the "differential variables"
            block_name = self.get_category_block_name(categ)
            set_name = self.get_category_set_name(categ)
            set_range = range(len(varlist))

            # Construct a set that indexes, eg, the "differential variables"
            category_set = Set(initialize=set_range)
            self.add_component(set_name, category_set)

            # Construct an IndexedBlock, each data object of which
            # will contain a single reference-to-timeslice of that
            # category, and with the corresponding custom ctype
            category_block = Block(category_set)
            self.add_component(block_name, category_block)

            # Don't want these blocks sent to any solver.
            category_block.deactivate()

            for i, var in enumerate(varlist):
                # Add reference-to-timeslices to new blocks:
                #
                # Create a new reference if var is not a reference or
                # it has the wrong ctype...
                # We may want to reconsider overriding the user's ctype...
                # We may also want to make sure these references are not
                # attached to anything. Otherwise we will fail here.
                ref = (
                    var
                    if var.is_reference() and var.ctype is ctype
                    else Reference(var, ctype=ctype)
                )
                category_block[i].add_component(var_name, ref)
                # These vars were created by the categorizer
                # and have custom ctypes.

    _vectors_name = "vectors"

    def _add_category_references(self):
        """Create a "time-indexed vector" for each category of variables."""
        category_dict = self.category_dict

        # Add a deactivated block to store all my `_NmpcVector`s
        # These be will vars, named by category, indexed by the index
        # into the list of that category and by time. E.g.
        # self.vectors.differential
        self.add_component(self._vectors_name, Block())
        self.vectors.deactivate()

        for categ in category_dict:
            ctype = CATEGORY_TYPE_MAP.get(categ, NmpcVar)
            # Get the block that holds this category of var,
            # and the name of the attribute that holds the
            # custom-ctype var (this attribute is the same
            # for all blocks).
            block_name = self.get_category_block_name(categ)
            var_name = self._var_name

            # Get a slice of the block, e.g. self.DIFFERENTIAL_BLOCK[:]
            block = getattr(self, block_name)
            _slice = block[:]
            # _slice = self.__getattribute__(block_name)[:]
            # Why does this work when self.__getattr__(block_name) does not?
            # __getattribute__ appears to work just fine...

            # Get a slice of the block and var, e.g.
            # self.DIFFERENTIAL_BLOCK[:].var[:]
            _slice = getattr(_slice, var_name)[:]

            # Add a reference to this slice to the `vectors` block.
            # This will be, e.g. `self.vectors.differential` and
            # can be accessed with its two indices, e.g.
            # `self.vectors.differential[i,t0]`
            # to get the "ith coordinate" of the vector of differential
            # variables at time t0.
            ref = Reference(_slice, ctype=_NmpcVector)
            self.vectors.add_component(
                categ.name.lower(),  # Lowercase of the enum name
                ref,
            )

    # Time is added in DynamicBlock.construct but this is nice if the user wants
    # to add time in a rule without a long messy line of super().__setattr__.
    def add_time(self):
        # Do this because I can't add a reference to a set
        super(BlockData, self).__setattr__("time", self.time)

    def set_sample_time(self, sample_time, tolerance=1e-8):
        """Validates and sets sample time"""
        self.validate_sample_time(sample_time, tolerance)
        self.sample_time = sample_time

    def validate_sample_time(self, sample_time, tolerance=1e-8):
        """Makes sure sample points, or integer multiple of sample time-offsets
        from time.first(), lie on finite element boundaries, and that the
        horizon of each model is an integer multiple of sample time. Assembles
        a list of sample points and a dictionary mapping sample points to the
        number of finite elements in the preceding sampling period, and adds
        them as attributes to _NMPC_NAMESPACE.

        Args:
            sample_time: Sample time to check
            tolerance: Tolerance within which time points must be integer
                       multiples of sample time

        """
        time = self.time
        horizon_length = time.last() - time.first()
        n_t = len(time)

        # TODO: This should probably be a DAE utility
        min_spacing = horizon_length
        for t in time:
            if t == time.first():
                continue
            prev = time.prev(t)
            if t - prev < min_spacing:
                min_spacing = t - prev
        # Sanity check:
        assert min_spacing > 0
        # Required so only one point can satisfy equality to tolerance
        if tolerance >= min_spacing / 2:
            raise ValueError(
                "ContinuousSet tolerance is larger than half the minimum "
                "spacing. An element of this set will not necessarily be "
                "unique within this tolerance."
            )

        off_by = abs(remainder(horizon_length, sample_time))
        if off_by > tolerance:
            raise ValueError(
                "Sampling time must be an integer divider of "
                "horizon length within tolerance %f" % tolerance
            )
        n_samples = round(horizon_length / sample_time)
        self.samples_per_horizon = n_samples

        finite_elements = time.get_finite_elements()
        fe_set = set(finite_elements)
        finite_element_indices = [i for i in range(1, n_t + 1) if time.at(i) in fe_set]
        sample_points = [time.first()]
        sample_indices = [1]  # Indices of sample points with in time set
        sample_no = 1
        fe_per = 0
        fe_per_sample_dict = {}
        for i, t in zip(finite_element_indices, finite_elements):
            if t == time.first():
                continue
            fe_per += 1
            time_since = t - time.first()
            sp = sample_no * sample_time
            diff = abs(sp - time_since)
            if diff < tolerance:
                sample_points.append(t)
                sample_indices.append(i)
                sample_no += 1
                fe_per_sample_dict[sample_no] = fe_per
                fe_per = 0
            if time_since > sp:
                raise ValueError(
                    "Could not find a time point for the %ith "
                    "sample point" % sample_no
                )
        assert len(sample_points) == n_samples + 1
        self.fe_per_sample = fe_per_sample_dict
        self.sample_points = sample_points
        self.sample_point_indices = sample_indices

    def initialize_sample_to_setpoint(
        self,
        sample_idx,
        ctype=(DiffVar, AlgVar, InputVar, DerivVar),
    ):
        """Set values to setpoint values for variables of the
        specified variable ctypes in the specified sample.
        """
        time = self.time
        sample_point_indices = self.sample_point_indices
        i_0 = sample_point_indices[sample_idx - 1]
        i_s = sample_point_indices[sample_idx]
        for var in self.component_objects(ctype):
            # `type(var)` is a subclass of `NmpcVar`, so I can
            # access the `setpoint` attribute.
            #
            # Would like:
            # var[t1:ts].set_value(var.setpoint)
            for i in range(i_0 + 1, i_s + 1):
                # Want to exclude first time point of sample,
                # but include last time point of sample.
                t = time.at(i)
                var[t].set_value(var.setpoint)

    def initialize_sample_to_initial(
        self,
        sample_idx,
        ctype=(DiffVar, AlgVar, DerivVar),
    ):
        """Set values to initial values for variables of the
        specified variable ctypes in the specified sample.
        """
        time = self.time
        sample_point_indices = self.sample_point_indices
        i_0 = sample_point_indices[sample_idx - 1]
        i_s = sample_point_indices[sample_idx]
        t0 = time.at(i_0)
        for var in self.component_objects(ctype):
            # Would be nice if I could use a slice with
            # start/stop indices to make this more concise.
            for i in range(i_0 + 1, i_s + 1):
                t = time.at(i)
                var[t].set_value(var[t0].value)

    def initialize_to_setpoint(
        self,
        ctype=(DiffVar, AlgVar, InputVar, DerivVar),
    ):
        """Sets values to setpoint values for specified variable
        ctypes for all time points.
        """
        # There should be negligible overhead to initializing
        # in many small loops as opposed to one big loop here.
        for i in range(len(self.sample_points)):
            self.initialize_sample_to_setpoint(i, ctype=ctype)

    def initialize_to_initial_conditions(
        self,
        ctype=(DiffVar, AlgVar, DerivVar),
    ):
        """Sets values to initial values for specified variable
        ctypes for all time points.
        """
        # There should be negligible overhead to initializing
        # in many small loops as opposed to one big loop here.
        for i in range(len(self.sample_points)):
            self.initialize_sample_to_initial(i, ctype=ctype)

    def initialize_by_solving_elements(self, solver, **kwargs):
        """Solve the square problem with fixed inputs in each
        of the time finite elements individually. This can be
        thought of as a time integration.
        """
        strip_var_bounds = kwargs.pop("strip_var_bounds", True)
        input_option = kwargs.pop("input_option", InputOption.CURRENT)
        config = self.CONFIG(kwargs)
        square_solve_context = SquareSolveContext(
            self,
            strip_var_bounds=strip_var_bounds,
            input_option=input_option,
        )
        model = self.mod
        time = self.time
        # There is a significant amount of overhead when calling
        # initialize_by_element_in_range multiple times, so
        # this method does not call `initialize_samples_by_element`
        # in a loop.
        if VC.DIFFERENTIAL in self.categories:
            time_linking_vars = self.differential_vars
        else:
            time_linking_vars = None
        with square_solve_context as sqs:
            initialize_by_element_in_range(
                model,
                time,
                time.first(),
                time.last(),
                dae_vars=self.dae_vars,
                # time_linking_vars=list(self.differential_vars[:]),
                time_linking_vars=time_linking_vars,
                outlvl=config.outlvl,
                solver=solver,
            )

    def initialize_samples_by_element(self, samples, solver, **kwargs):
        """Solve the square problem with fixed inputs for the specified samples"""
        # TODO: ConfigBlock for this class
        strip_var_bounds = kwargs.pop("strip_var_bounds", True)
        input_option = kwargs.pop("input_option", InputOption.CURRENT)
        config = self.CONFIG(kwargs)

        if type(samples) not in {list, tuple}:
            samples = (samples,)

        # Create a context manager that will temporarily strip bounds
        # and fix inputs, preparing the model for a "square solve."
        square_solve_context = SquareSolveContext(
            self,
            samples=samples,
            strip_var_bounds=strip_var_bounds,
            input_option=input_option,
        )
        sample_points = self.sample_points
        model = self.mod
        time = self.time
        with square_solve_context as sqs:
            for s in samples:
                t0 = sample_points[s - 1]
                t1 = sample_points[s]
                # Really I would like an `ElementInitializer` context manager
                # class that deactivates the model once, then allows me to
                # activate the elements I want to solve one at a time.
                # This would allow me to not repeat so much work when
                # initializing multiple samples.
                initialize_by_element_in_range(
                    model,
                    time,
                    t0,
                    t1,
                    dae_vars=self.dae_vars,
                    time_linking_vars=list(self.differential_vars[:]),
                    outlvl=config.outlvl,
                    solver=solver,
                )

    def set_variance(self, variance_list):
        """Set variance for corresponding NmpcVars to the values provided

        Arguments:
            variance_list: List of vardata, value tuples. The vardatas
                           correspond to time-indexed references, and values
                           are the variances.
        """
        t0 = self.time.first()
        variance_map = ComponentMap(variance_list)
        for var, val in variance_list:
            nmpc_var = self.vardata_map[var]
            nmpc_var.variance = val
        # MeasurementVars will not have their variance set since they are
        # not mapped to in vardata_map
        for var in self.component_objects(MeasuredVar):
            # component_objects is fine here because we don't need
            # to access measurements in any particular order.
            if var[t0] in variance_map:
                var.variance = variance_map[var[t0]]

    def generate_inputs_at_time(self, t):
        if VC.INPUT in self.categories:
            for val in self.vectors.input[:, t].value:
                yield val
        else:
            raise RuntimeError(
                "Trying to generate inputs but no input " "category has been specified."
            )

    def generate_measurements_at_time(self, t):
        if VC.MEASUREMENT in self.categories:
            for val in self.vectors.measurement[:, t].value:
                yield val
        else:
            raise RuntimeError(
                "Trying to generate measurements but no measurement "
                "category has been specified."
            )

    def inject_inputs(self, inputs):
        # To simulate computational delay, this function would
        # need an argument for the start time of inputs.
        if VC.INPUT in self.categories:
            for var, val in zip(self.INPUT_BLOCK[:].var, inputs):
                # Would like:
                # self.vectors.input[:,:].fix(inputs)
                # This is an example of setting a matrix from a vector.
                # Could even aspire towards:
                # self.vectors.input[:,t0:t1].fix(inputs)
                var[:].fix(val)
        else:
            raise RuntimeError(
                "Trying to set input values but no input "
                "category has been specified."
            )

    def load_measurements(self, measured):
        t0 = self.time.first()
        if VC.MEASUREMENT in self.categories:
            # Want: self.measured_vars[:,t0].fix(measured)
            for var, val in zip(self.MEASUREMENT_BLOCK[:].var, measured):
                var[t0].fix(val)
        else:
            raise RuntimeError(
                "Trying to set measurement values but no measurement "
                "category has been specified."
            )

    def advance_by_time(
        self,
        t_shift,
        ctype=(DiffVar, DerivVar, AlgVar, InputVar, FixedVar),
        # Fixed variables are included as I expect disturbances
        # should shift in time as well.
        tolerance=1e-8,
    ):
        """Set values for the variables of the specified ctypes
        to their values `t_shift` in the future.
        """
        time = self.time
        # The outer loop is over time so we don't have to call
        # `find_nearest_index` for every variable.
        # I am assuming that `find_nearest_index` is slower than
        # accessing `component_objects`
        for t in time:
            ts = t + t_shift
            idx = time.find_nearest_index(ts, tolerance)
            if idx is None:
                # t + sample_time is outside the model's "horizon"
                continue
            ts = time.at(idx)
            for var in self.component_objects(ctype):
                var[t].set_value(var[ts].value)

    def advance_one_sample(
        self,
        ctype=(DiffVar, DerivVar, AlgVar, InputVar, FixedVar),
        tolerance=1e-8,
    ):
        """Set values for the variables of the specified ctypes
        to their values one sample time in the future.
        """
        sample_time = self.sample_time
        self.advance_by_time(
            sample_time,
            ctype=ctype,
            tolerance=tolerance,
        )

    def generate_time_in_sample(
        self,
        ts,
        t0=None,
        include_t0=False,
        tolerance=1e-8,
    ):
        """Generate time points between the provided time point
        and one sample time in the past.
        """
        # TODO: Need to address the question of whether I want users
        # passing around time points or the integer index of samples.
        time = self.time
        idx_s = time.find_nearest_index(ts, tolerance=tolerance)
        ts = time.at(idx_s)
        if t0 is None:
            t0 = ts - self.sample_time
        idx_0 = time.find_nearest_index(t0, tolerance=tolerance)
        idx_start = idx_0 if include_t0 else idx_0 + 1
        for i in range(idx_start, idx_s + 1):
            # Don't want to include first point in sample
            yield time.at(i)

    def get_data_from_sample(
        self,
        ts,
        variables=(
            VC.DIFFERENTIAL,
            VC.INPUT,
        ),
        tolerance=1e-8,
        include_t0=False,
    ):
        """Creates an `OrderedDict` that maps the time-indexed reference
        of each variable provided to a list of its values over the sample
        preceding the specified time point.
        """
        time = self.time
        sample_time = self.sample_time
        category_dict = self.category_dict
        vardata_map = self.vardata_map

        data = OrderedDict()
        queue = list(variables)
        for var in queue:
            if type(var) is VC:
                category = var
                varlist = category_dict[category]
                queue.extend(var[ts] for var in varlist)
                continue
            _slice = vardata_map[var]
            cuid = ComponentUID(_slice.referent)
            if include_t0:
                i0 = time.find_nearest_index(ts - sample_time, tolerance=tolerance)
                t0 = time.at(i0)
                data[cuid] = [_slice[t0].value]
            else:
                data[cuid] = []
            data[cuid].extend(
                _slice[t].value
                for t in self.generate_time_in_sample(ts, tolerance=tolerance)
            )

        return data

    def add_ipopt_suffixes(self):
        """Adds suffixes for communicating dual variables with IPOPT"""
        # Maybe there should be some helper class to do solver-specific
        # stuff like this...
        self.ipopt_zL_out = Suffix(direction=Suffix.IMPORT)
        self.ipopt_zU_out = Suffix(direction=Suffix.IMPORT)

        self.ipopt_zL_in = Suffix(direction=Suffix.EXPORT)
        self.ipopt_zU_in = Suffix(direction=Suffix.EXPORT)

        self.dual = Suffix(direction=Suffix.IMPORT_EXPORT)

    def update_ipopt_multipliers(self):
        self.ipopt_zL_in.update(self.ipopt_zL_out)
        self.ipopt_zU_in.update(self.ipopt_zU_out)

    def advance_ipopt_multipliers(
        self,
        t_shift,
        ctype=(
            DiffVar,
            AlgVar,
            InputVar,
        ),
        tolerance=1e-8,
    ):
        """Set the values of bound multipliers to the corresponding
        values a time `t_shift` in the future.
        """
        zL = self.ipopt_zL_in
        zU = self.ipopt_zU_in
        time = self.time
        # The outer loop is over time so we don't have to call
        # `find_nearest_index` for every variable.
        # I am assuming that `find_nearest_index` is slower than
        # accessing `component_objects`
        for t in time:
            ts = t + t_shift
            idx = time.find_nearest_index(ts, tolerance)
            if idx is None:
                # t + sample_time is outside the model's "horizon"
                continue
            ts = time.at(idx)
            for var in self.component_objects(ctype):
                if var[t] in zL and var[ts] in zL:
                    zL[var[t]] = zL[var[ts]]
                if var[t] in zU and var[ts] in zU:
                    zU[var[t]] = zU[var[ts]]

    def advance_ipopt_multipliers_one_sample(
        self,
        ctype=(
            DiffVar,
            AlgVar,
            InputVar,
        ),
        tolerance=1e-8,
    ):
        """Set the values of bound multipliers to the corresponding
        values one sample time in the future.
        """
        sample_time = self.sample_time
        self.advance_ipopt_multipliers(
            sample_time,
            ctype=ctype,
            tolerance=tolerance,
        )


class DynamicBlock(Block):
    """This is the DynamicBlock class that the user will instantiate."""

    _ComponentDataClass = _DynamicBlockData

    def __new__(cls, *args, **kwds):
        # Decide what class to allocate
        if cls != DynamicBlock:
            target_cls = cls
        elif not args or (args[0] is UnindexedComponent_set and len(args) == 1):
            target_cls = SimpleDynamicBlock
        else:
            target_cls = IndexedDynamicBlock
        return super(DynamicBlock, cls).__new__(target_cls)

    def __init__(self, *args, **kwds):
        # This will get called regardless of what class we are instantiating
        self._init_model = Initializer(kwds.pop("model", None))
        self._init_time = Initializer(
            kwds.pop("time", None), treat_sequences_as_mappings=False
        )
        self._init_inputs = Initializer(
            kwds.pop("inputs", None), treat_sequences_as_mappings=False
        )
        self._init_measurements = Initializer(
            kwds.pop("measurements", None), treat_sequences_as_mappings=False
        )
        self._init_category_dict = Initializer(
            kwds.pop("category_dict", None), treat_sequences_as_mappings=False
        )
        Block.__init__(self, *args, **kwds)

    def _getitem_when_not_present(self, idx):
        block = super(DynamicBlock, self)._getitem_when_not_present(idx)
        parent = self.parent_block()

        if self._init_model is not None:
            block.mod = self._init_model(parent, idx)

        if self._init_time is not None:
            super(BlockData, block).__setattr__("time", self._init_time(parent, idx))

        if self._init_inputs is not None:
            block._inputs = self._init_inputs(parent, idx)

        if self._init_measurements is not None:
            block._measurements = self._init_measurements(parent, idx)

        if self._init_category_dict is not None:
            block._category_dict = self._init_category_dict(parent, idx)
        else:
            block._category_dict = None

        block._construct()


class SimpleDynamicBlock(_DynamicBlockData, DynamicBlock):
    def __init__(self, *args, **kwds):
        _DynamicBlockData.__init__(self, component=self)
        DynamicBlock.__init__(self, *args, **kwds)

    # Pick up the display() from Block and not BlockData
    display = DynamicBlock.display


class IndexedDynamicBlock(DynamicBlock):
    def __init__(self, *args, **kwargs):
        DynamicBlock.__init__(self, *args, **kwargs)


class SquareSolveContext(object):
    """
    Utility class to prepare DynamicBlock for a square solve.
    """

    def __init__(
        self,
        dynamic_block,
        samples=None,
        strip_var_bounds=True,
        input_option=InputOption.CURRENT,
    ):
        """
        Parameters
        ----------
            dynamic_block: A _DynamicBlockData object
            samples: A list of integers corresponding to the samples that
                     will be solved

        """
        self.block = dynamic_block
        self.samples = samples
        self.strip_var_bounds = strip_var_bounds
        self.input_option = input_option

        # Get indices for the time points where we need to fix inputs.
        if samples is None:
            # Assume we need to fix inputs for all non-initial time
            self.time_indices = range(2, len(dynamic_block.time) + 1)
        else:
            # Get the indices of all time points in the samples specified
            sample_points = dynamic_block.sample_points
            time = dynamic_block.time
            time_indices = []
            already_visited = set()
            for s in samples:
                # s is an integer in range(1, len(sample_points)+1)
                # the ith sample is the interval (ts_{i-1}, ts_i]
                t0 = sample_points[s - 1]
                ts = sample_points[s]
                idx_0 = time.find_nearest_index(t0)
                idx_s = time.find_nearest_index(ts)
                for i in range(idx_0 + 1, idx_s + 1):  # Pyomo sets are 1-indexed
                    if i not in already_visited:
                        # Want to make sure each index gets added at most
                        # once in the case of repeated samples...
                        time_indices.append(i)
                        already_visited.add(i)
            self.time_indices = time_indices

    def __enter__(self):
        # Strip bounds:
        if self.strip_var_bounds:
            self.strip_bounds = TransformationFactory("contrib.strip_var_bounds")
            self.strip_bounds.apply_to(self.block.mod, reversible=True)

        # Fix inputs:
        time = self.block.time
        t0 = time.first()
        time_indices = self.time_indices
        input_vars = self.block.input_vars
        input_option = self.input_option
        if input_option is InputOption.CURRENT:
            input_vals = [None for _ in input_vars]
        elif input_option is InputOption.INITIAL:
            input_vals = [v[t0] for v in input_vars]
        elif input_option is InputOption.SETPOINT:
            input_vals = [v.setpoint for v in input_vars]
        else:
            raise NotImplementedError("Unrecognized input option")

        for var, val in zip(input_vars, input_vals):
            for i in time_indices:
                t = time.at(i)
                if val is None:
                    var[t].fix()
                else:
                    var[t].fix(val)

        # Model should now be square and bound-free
        return self

    def __exit__(self, ex_type, ex_val, ex_tb):
        # Unfix inputs:
        time = self.block.time
        time_indices = self.time_indices
        input_vars = self.block.input_vars
        for var in input_vars:
            for i in time_indices:
                t = time.at(i)
                var[t].unfix()

        if self.strip_var_bounds:
            self.strip_bounds.revert(self.block.mod)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Methods for cubic equations of state.

Currently only supports liquid and vapor phases
"""
# TODO: Pylint complains about variables with _x names as they are built by other classes
# pylint: disable=protected-access
# pylint: disable=missing-function-docstring

from enum import Enum
from copy import deepcopy

from pyomo.environ import (
    Constraint,
    exp,
    Expression,
    log,
    Reals,
    sqrt,
    Var,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In
from idaes.models.properties.modular_properties.base.utility import (
    get_method,
    get_component_object as cobj,
)

from idaes.core.util.math import safe_log

import idaes.logger as idaeslog
from idaes.core.util.exceptions import (
    BurntToast,
    ConfigurationError,
    PropertyNotSupportedError,
)

# cubic_roots_available is used elsewhere
# pylint: disable=W0611
from idaes.models.properties.modular_properties.eos.ceos_common import (
    EoS_param,
    cubic_roots_available,
    CubicThermoExpressions,
    CubicType,
)
from .eos_base import EoSBase

# Set up logger
_log = idaeslog.getLogger(__name__)


"""
References:
[1]. Poling, B.E., Prausnitz, J.M. and O’connell, J.P., 2001.
     Properties of gases and liquids. McGraw-Hill Education.

[2]. Trujillo, M.F., O'Rourke, P. and Torres, D., 2002.
     Generalizing the Thermodynamics State Relationships in KIVA-3V
     https://www.osti.gov/servlets/purl/809947 (Last accessed: 08/13/2021)
"""


class MixingRuleA(Enum):
    """Enum for supported rules for calculating am"""

    # Rule to calculate am for cubic equations of state
    default = 0


class MixingRuleB(Enum):
    """Enum for supported rules for calculating bm"""

    # Rule to calculate bm for cubic equations of state
    default = 0


# Value for smoothing epsilon for SafeLog when used
eps_SL = 1e-8

CubicConfig = ConfigBlock()
CubicConfig.declare(
    "type",
    ConfigValue(
        domain=In(CubicType),
        description="Equation of state to use",
        doc="Enum indicating type of cubic equation of state to use.",
    ),
)
CubicConfig.declare(
    "mixing_rule_a",
    ConfigValue(
        default=MixingRuleA.default,
        domain=In(MixingRuleA),
        description="Mixing rule for parameter am for cubic EoS",
        doc="Enum indicating type of mixing rule for parameter am to use.",
    ),
)
CubicConfig.declare(
    "mixing_rule_b",
    ConfigValue(
        default=MixingRuleB.default,
        domain=In(MixingRuleB),
        description="Mixing rule for parameter bm for cubic EoS",
        doc="Enum indicating type of mixing rule for parameter bm to use.",
    ),
)


class Cubic(EoSBase):
    """Class for constructing modular cubic EoS properties"""

    @staticmethod
    def common(b, pobj):
        # TODO: determine if Henry's Law applies to Cubic EoS systems
        # For now, raise an exception if found
        # Follow on questions:
        # If Henry's law is used for a component, how does that effect
        # calculating A, B and phi?
        for j in b.component_list:
            cobj = b.params.get_component(j)
            if (
                cobj.config.henry_component is not None
                and pobj.local_name in cobj.config.henry_component
            ):
                raise PropertyNotSupportedError(
                    "{} Cubic equations of state do not support Henry's "
                    "components [{}, {}].".format(b.name, pobj.local_name, j)
                )

        ctype = pobj._cubic_type
        cname = pobj.config.equation_of_state_options["type"].name
        mixing_rule_a = pobj._mixing_rule_a
        mixing_rule_b = pobj._mixing_rule_b

        if hasattr(b, cname + "_fw"):
            # Common components already constructed by previous phase
            return

        # Create expressions for coefficients
        def rule_fw(m, j):
            func_fw = getattr(m.params, cname + "_func_fw")
            cobj = m.params.get_component(j)
            return func_fw(cobj)

        b.add_component(
            cname + "_fw",
            Expression(b.component_list, rule=rule_fw, doc="EoS S factor"),
        )

        def rule_a_crit(m, j):
            cobj = m.params.get_component(j)
            return EoS_param[ctype]["omegaA"] * (
                (Cubic.gas_constant(b) * cobj.temperature_crit) ** 2
                / cobj.pressure_crit
            )

        b.add_component(
            cname + "_a_crit",
            Expression(
                b.component_list,
                rule=rule_a_crit,
                doc="Component a coefficient at T_crit",
            ),
        )

        def rule_a(m, j):
            cobj = m.params.get_component(j)
            fw = getattr(m, cname + "_fw")[j]
            ac = getattr(m, cname + "_a_crit")[j]
            func_alpha = getattr(m.params, cname + "_func_alpha")

            return ac * func_alpha(m.temperature, fw, cobj)

        b.add_component(
            cname + "_a",
            Expression(b.component_list, rule=rule_a, doc="Component a coefficient"),
        )

        def rule_da_dT(m, j):
            cobj = m.params.get_component(j)
            fw = getattr(m, cname + "_fw")[j]
            ac = getattr(m, cname + "_a_crit")[j]
            func_dalpha_dT = getattr(m.params, cname + "_func_dalpha_dT")

            return ac * func_dalpha_dT(m.temperature, fw, cobj)

        b.add_component(
            cname + "_da_dT",
            Expression(
                b.component_list,
                rule=rule_da_dT,
                doc="Temperature derivative of component a",
            ),
        )

        def rule_d2a_dT2(m, j):
            cobj = m.params.get_component(j)
            fw = getattr(m, cname + "_fw")[j]
            ac = getattr(m, cname + "_a_crit")[j]
            func_d2alpha_dT2 = getattr(m.params, cname + "_func_d2alpha_dT2")

            return ac * func_d2alpha_dT2(m.temperature, fw, cobj)

        b.add_component(
            cname + "_d2a_dT2",
            Expression(
                b.component_list,
                rule=rule_d2a_dT2,
                doc="Second temperature derivative" "of component a",
            ),
        )

        def func_b(m, j):
            cobj = m.params.get_component(j)
            return (
                EoS_param[ctype]["coeff_b"]
                * Cubic.gas_constant(b)
                * cobj.temperature_crit
                / cobj.pressure_crit
            )

        b.add_component(
            cname + "_b",
            Expression(b.component_list, rule=func_b, doc="Component b coefficient"),
        )

        if mixing_rule_a == MixingRuleA.default:

            def rule_am(m, p):
                a = getattr(m, cname + "_a")
                return rule_am_default(m, cname, a, p)

            b.add_component(cname + "_am", Expression(b.phase_list, rule=rule_am))

            def rule_daij_dT(m, i, j):
                a = getattr(m, cname + "_a")
                da_dT = getattr(m, cname + "_da_dT")
                k = getattr(m.params, cname + "_kappa")

                # Include temperature derivative of k for future extension
                dk_ij_dT = 0

                return sqrt(a[i] * a[j]) * (
                    -dk_ij_dT + (1 - k[i, j]) / 2 * (da_dT[i] / a[i] + da_dT[j] / a[j])
                )

            b.add_component(
                cname + "_daij_dT",
                Expression(b.component_list, b.component_list, rule=rule_daij_dT),
            )

            def rule_dam_dT(m, p):
                daij_dT = getattr(m, cname + "_daij_dT")
                return sum(
                    sum(
                        m.mole_frac_phase_comp[p, i]
                        * m.mole_frac_phase_comp[p, j]
                        * daij_dT[i, j]
                        for j in m.components_in_phase(p)
                    )
                    for i in m.components_in_phase(p)
                )

            b.add_component(
                cname + "_dam_dT", Expression(b.phase_list, rule=rule_dam_dT)
            )

            def rule_d2am_dT2(m, p):
                k = getattr(m.params, cname + "_kappa")
                a = getattr(m, cname + "_a")
                da_dT = getattr(m, cname + "_da_dT")
                d2a_dT2 = getattr(m, cname + "_d2a_dT2")

                # Initialize loop variable
                d2am_dT2 = None

                for i in m.components_in_phase(p):
                    for j in m.components_in_phase(p):
                        d2aij_dT2 = sqrt(a[i] * a[j]) * (
                            # -d2k_dT2  # TODO: Placeholder for if temperature dependent k is needed
                            # - dk_dT * (da_dT[i] / a[i] + da_dT[j] / a[j])
                            +(1 - k[i, j])
                            / 2
                            * (
                                d2a_dT2[i] / a[i]
                                + d2a_dT2[j] / a[j]
                                - 1 / 2 * (da_dT[i] / a[i] - da_dT[j] / a[j]) ** 2
                            )
                        )
                        d2am_dT2_term = (
                            m.mole_frac_phase_comp[p, i]
                            * m.mole_frac_phase_comp[p, j]
                            * d2aij_dT2
                        )
                        if d2am_dT2 is None:
                            d2am_dT2 = d2am_dT2_term
                        else:
                            d2am_dT2 += d2am_dT2_term
                return d2am_dT2

            b.add_component(
                cname + "_d2am_dT2", Expression(b.phase_list, rule=rule_d2am_dT2)
            )

            def rule_delta(m, p, i):
                # See pg. 145 in Properties of Gases and Liquids
                a = getattr(m, cname + "_a")
                am = getattr(m, cname + "_am")
                kappa = getattr(m.params, cname + "_kappa")
                return (
                    2
                    * sqrt(a[i])
                    / am[p]
                    * sum(
                        m.mole_frac_phase_comp[p, j] * sqrt(a[j]) * (1 - kappa[i, j])
                        for j in b.components_in_phase(p)
                    )
                )

            b.add_component(
                cname + "_delta", Expression(b.phase_component_set, rule=rule_delta)
            )

        else:
            raise ConfigurationError(
                "{} Unrecognized option for Equation of State "
                "mixing_rule_a: {}. Must be an instance of MixingRuleA "
                "Enum.".format(b.name, mixing_rule_a)
            )

        if mixing_rule_b == MixingRuleB.default:

            def rule_bm(m, p):
                b = getattr(m, cname + "_b")
                return rule_bm_default(m, b, p)

            b.add_component(cname + "_bm", Expression(b.phase_list, rule=rule_bm))
        else:
            raise ConfigurationError(
                "{} Unrecognized option for Equation of State "
                "mixing_rule_b: {}. Must be an instance of MixingRuleB "
                "Enum.".format(b.name, mixing_rule_b)
            )

        def rule_A(m, p):
            am = getattr(m, cname + "_am")
            return am[p] * m.pressure / (Cubic.gas_constant(b) * m.temperature) ** 2

        b.add_component(cname + "_A", Expression(b.phase_list, rule=rule_A))

        def rule_B(m, p):
            bm = getattr(m, cname + "_bm")
            return bm[p] * m.pressure / (Cubic.gas_constant(b) * m.temperature)

        b.add_component(cname + "_B", Expression(b.phase_list, rule=rule_B))

        # Add components at equilibrium state if required
        if b.params.config.phases_in_equilibrium is not None and (
            not b.config.defined_state or b.always_flash
        ):

            def func_a_eq(m, p1, p2, j):
                cobj = m.params.get_component(j)
                fw = getattr(m, cname + "_fw")[j]
                ac = getattr(m, cname + "_a_crit")[j]
                func_alpha = getattr(m.params, cname + "_func_alpha")

                return ac * func_alpha(m._teq[p1, p2], fw, cobj)

            b.add_component(
                "_" + cname + "_a_eq",
                Expression(
                    b.params._pe_pairs,
                    b.component_list,
                    rule=func_a_eq,
                    doc="Component a coefficient at Teq",
                ),
            )

            def rule_am_eq(m, p1, p2, p3):
                try:
                    rule = m.params.get_phase(p3).config.equation_of_state_options[
                        "mixing_rule_a"
                    ]
                except (KeyError, TypeError):
                    rule = MixingRuleA.default

                a = getattr(m, "_" + cname + "_a_eq")
                if rule == MixingRuleA.default:
                    return rule_am_default(m, cname, a, p3, (p1, p2))
                else:
                    raise ConfigurationError(
                        "{} Unrecognized option for Equation of State "
                        "mixing_rule_a: {}. Must be an instance of MixingRuleA "
                        "Enum.".format(m.name, rule)
                    )

            b.add_component(
                "_" + cname + "_am_eq",
                Expression(b.params._pe_pairs, b.phase_list, rule=rule_am_eq),
            )

            def rule_A_eq(m, p1, p2, p3):
                am_eq = getattr(m, "_" + cname + "_am_eq")
                return (
                    am_eq[p1, p2, p3]
                    * m.pressure
                    / (Cubic.gas_constant(b) * m._teq[p1, p2]) ** 2
                )

            b.add_component(
                "_" + cname + "_A_eq",
                Expression(b.params._pe_pairs, b.phase_list, rule=rule_A_eq),
            )

            def rule_B_eq(m, p1, p2, p3):
                bm = getattr(m, cname + "_bm")
                return bm[p3] * m.pressure / (Cubic.gas_constant(b) * m._teq[p1, p2])

            b.add_component(
                "_" + cname + "_B_eq",
                Expression(b.params._pe_pairs, b.phase_list, rule=rule_B_eq),
            )

            def rule_delta_eq(m, p1, p2, p3, i):
                # See pg. 145 in Properties of Gases and Liquids
                a = getattr(m, "_" + cname + "_a_eq")
                am = getattr(m, "_" + cname + "_am_eq")
                kappa = getattr(m.params, cname + "_kappa")
                return (
                    2
                    * sqrt(a[p1, p2, i])
                    / am[p1, p2, p3]
                    * sum(
                        m.mole_frac_phase_comp[p3, j]
                        * sqrt(a[p1, p2, j])
                        * (1 - kappa[i, j])
                        for j in m.components_in_phase(p3)
                    )
                )

            b.add_component(
                "_" + cname + "_delta_eq",
                Expression(
                    b.params._pe_pairs, b.phase_component_set, rule=rule_delta_eq
                ),
            )

    @staticmethod
    def calculate_scaling_factors(b, pobj):
        pass

    @staticmethod
    def build_parameters(b):
        param_block = b.parent_block()
        if not (b.is_vapor_phase() or b.is_liquid_phase()):
            raise PropertyNotSupportedError(
                "{} received unrecognized phase "
                "name {}. Cubic equation of state supports only Vap and Liq "
                "phases.".format(param_block.name, b)
            )

        if b.config.equation_of_state_options["type"] not in set(
            item for item in CubicType
        ):
            raise ConfigurationError(
                "{} Unrecognized option for equation of "
                "state type: {}. Must be an instance of CubicType "
                "Enum.".format(b.name, b.config.equation_of_state_options["type"])
            )

        ctype = b.config.equation_of_state_options["type"]
        b._cubic_type = ctype
        cname = ctype.name

        # Check to see if ConfigBlock was created by previous phase
        if hasattr(param_block, cname + "_eos_options"):
            ConfigBlock = getattr(param_block, cname + "_eos_options")
            for key, value in b.config.equation_of_state_options.items():
                if ConfigBlock[key] != value:
                    raise ConfigurationError(
                        "In {}, different {} equation of "
                        "state options for {} are set in different phases, which is "
                        "not supported.".format(b.name, cname, key)
                    )
            # Once the options have been validated, we don't have anything
            # left to do
            mixing_rule_a = ConfigBlock["mixing_rule_a"]
            mixing_rule_b = ConfigBlock["mixing_rule_b"]
            b._mixing_rule_a = mixing_rule_a
            b._mixing_rule_b = mixing_rule_b
            return

        setattr(param_block, cname + "_eos_options", deepcopy(CubicConfig))
        ConfigBlock = getattr(param_block, cname + "_eos_options")
        ConfigBlock.set_value(b.config.equation_of_state_options)

        mixing_rule_a = ConfigBlock["mixing_rule_a"]
        mixing_rule_b = ConfigBlock["mixing_rule_b"]
        b._mixing_rule_a = mixing_rule_a
        b._mixing_rule_b = mixing_rule_b

        kappa_data = param_block.config.parameter_data[cname + "_kappa"]
        param_block.add_component(
            cname + "_kappa",
            Var(
                param_block.component_list,
                param_block.component_list,
                within=Reals,
                initialize=kappa_data,
                doc=cname + " binary interaction parameters",
                units=None,
            ),
        )

        if b._cubic_type == CubicType.PR:
            func_fw = func_fw_PR
        elif b._cubic_type == CubicType.SRK:
            func_fw = func_fw_SRK
        else:
            raise BurntToast(
                "{} received unrecognized cubic type. This should "
                "never happen, so please contact the IDAES developers "
                "with this bug.".format(b.name)
            )
        setattr(param_block, cname + "_func_fw", func_fw)
        setattr(param_block, cname + "_func_alpha", func_alpha_soave)
        setattr(param_block, cname + "_func_dalpha_dT", func_dalpha_dT_soave)
        setattr(param_block, cname + "_func_d2alpha_dT2", func_d2alpha_dT2_soave)

    @staticmethod
    def compress_fact_phase(b, p):
        """
        Compressibility factor
        """
        pobj = b.params.get_phase(p)
        cname = pobj._cubic_type.name
        A = getattr(b, cname + "_A")
        B = getattr(b, cname + "_B")

        expr_write = CubicThermoExpressions(b)
        if pobj.is_vapor_phase():
            return expr_write.z_vap(eos=pobj._cubic_type, A=A[p], B=B[p])
        elif pobj.is_liquid_phase():
            return expr_write.z_liq(eos=pobj._cubic_type, A=A[p], B=B[p])
        raise BurntToast(
            "{} non-vapor or liquid phase called for cubic "
            "EoS compressability factor. This should never "
            "happen, so please contact the IDAES developers "
            "with this bug.".format(b.name)
        )

    @staticmethod
    def cp_mass_phase(blk, p):
        """
        Phase mass-specific heat capacity at constant pressure
        """
        return blk.cp_mol_phase[p] / blk.mw_phase[p]

    @staticmethod
    def cp_mol_phase(blk, p):
        """
        Phase molar heat capacity at constant pressure
        """
        pobj = blk.params.get_phase(p)
        cname = pobj._cubic_type.name

        am = getattr(blk, cname + "_am")[p]
        bm = getattr(blk, cname + "_bm")[p]
        B = getattr(blk, cname + "_B")[p]

        dam_dT = getattr(blk, cname + "_dam_dT")[p]
        d2am_dT2 = getattr(blk, cname + "_d2am_dT2")[p]

        T = blk.temperature
        R = Cubic.gas_constant(blk)
        Z = blk.compress_fact_phase[p]
        dZdT = _dZ_dT(blk, p)

        EoS_u = EoS_param[pobj._cubic_type]["u"]
        EoS_w = EoS_param[pobj._cubic_type]["w"]
        EoS_p = sqrt(EoS_u**2 - 4 * EoS_w)

        expression1 = 2 * Z + (EoS_u + EoS_p) * B
        expression2 = 2 * Z + (EoS_u - EoS_p) * B
        expression3 = B * (dZdT + Z / T) / (Z**2 + Z * EoS_u * B + EoS_w * B**2)

        cp_ideal_gas = sum(
            blk.mole_frac_phase_comp[p, j]
            * get_method(blk, "cp_mol_ig_comp", j)(blk, cobj(blk, j), T)
            for j in blk.components_in_phase(p)
        )

        # Derived from the relations in Chapter 6 of [1]
        cp_departure = (
            R * (T * dZdT + Z - 1)
            + (T * d2am_dT2 / (EoS_p * bm))
            * safe_log(expression1 / expression2, eps=eps_SL)
            + ((am - T * dam_dT) * expression3 / bm)
        )

        return cp_ideal_gas + cp_departure

    @staticmethod
    def cv_mass_phase(blk, p):
        """
        Phase mass-specific heat capacity at constant volume
        """
        return blk.cv_mol_phase[p] / blk.mw_phase[p]

    @staticmethod
    def cv_mol_phase(blk, p):
        """
        Phase molar heat capacity at constant volume
        """
        pobj = blk.params.get_phase(p)
        cname = pobj._cubic_type.name
        am = getattr(blk, cname + "_am")[p]
        bm = getattr(blk, cname + "_bm")[p]
        cp = blk.cp_mol_phase[p]
        V = 1 / blk.dens_mol_phase[p]
        dam_dT = getattr(blk, cname + "_dam_dT")[p]

        EoS_u = EoS_param[pobj._cubic_type]["u"]
        EoS_w = EoS_param[pobj._cubic_type]["w"]

        dPdV = -((Cubic.gas_constant(blk) * blk.temperature) / (V - bm) ** 2) + (
            am * (2 * V + EoS_u * bm) / (V**2 + EoS_u * bm * V + EoS_w * bm**2) ** 2
        )

        dPdT = (Cubic.gas_constant(blk) / (V - bm)) - (
            1 / (V**2 + EoS_u * bm * V + EoS_w * bm**2)
        ) * dam_dT

        # See Chapter 6 in [1]
        return cp + blk.temperature * dPdT**2 / dPdV

    @staticmethod
    def dens_mass_phase(b, p):
        """
        Phase density (mass basis)
        """
        return b.dens_mol_phase[p] * b.mw_phase[p]

    @staticmethod
    def dens_mol_phase(b, p):
        """
        Phase density (mole basis)
        """
        return b.pressure / (
            Cubic.gas_constant(b) * b.temperature * b.compress_fact_phase[p]
        )

    @staticmethod
    def energy_internal_mol_phase(blk, p):
        """
        Phase molar internal energy
        """
        pobj = blk.params.get_phase(p)

        cname = pobj._cubic_type.name
        am = getattr(blk, cname + "_am")[p]
        bm = getattr(blk, cname + "_bm")[p]
        B = getattr(blk, cname + "_B")[p]
        dam_dT = getattr(blk, cname + "_dam_dT")[p]
        Z = blk.compress_fact_phase[p]

        EoS_u = EoS_param[pobj._cubic_type]["u"]
        EoS_w = EoS_param[pobj._cubic_type]["w"]
        EoS_p = sqrt(EoS_u**2 - 4 * EoS_w)

        # Derived from equation on pg. 120 in Properties of Gases and Liquids
        # Departure function for U is similar to H minus the RT(Z-1) term
        return (
            (blk.temperature * dam_dT - am)
            * safe_log(
                (2 * Z + B * (EoS_u + EoS_p)) / (2 * Z + B * (EoS_u - EoS_p)),
                eps=eps_SL,
            )
        ) / (bm * EoS_p) + sum(
            blk.mole_frac_phase_comp[p, j]
            * EoSBase.energy_internal_mol_ig_comp_pure(blk, j)
            for j in blk.components_in_phase(p)
        )

    @staticmethod
    def energy_internal_mol_phase_comp(blk, p, j):
        """
        Phase partial molar internal energy
        """
        return (
            blk.enth_mol_phase_comp[p, j] - blk.pressure * blk.vol_mol_phase_comp[p, j]
        )

    @staticmethod
    def enth_mol_phase(blk, p):
        """
        Phase molar enthalpy
        """
        pobj = blk.params.get_phase(p)

        cname = pobj._cubic_type.name
        am = getattr(blk, cname + "_am")[p]
        bm = getattr(blk, cname + "_bm")[p]
        B = getattr(blk, cname + "_B")[p]
        dam_dT = getattr(blk, cname + "_dam_dT")[p]
        Z = blk.compress_fact_phase[p]
        R = Cubic.gas_constant(blk)
        T = blk.temperature

        EoS_u = EoS_param[pobj._cubic_type]["u"]
        EoS_w = EoS_param[pobj._cubic_type]["w"]
        EoS_p = sqrt(EoS_u**2 - 4 * EoS_w)

        enth_ideal = sum(
            blk.mole_frac_phase_comp[p, j]
            * get_method(blk, "enth_mol_ig_comp", j)(blk, cobj(blk, j), blk.temperature)
            for j in blk.components_in_phase(p)
        )

        # Derived from equation on pg. 120 in Properties of Gases and Liquids
        enth_departure = R * T * (Z - 1) + (T * dam_dT - am) / (bm * EoS_p) * safe_log(
            (2 * Z + B * (EoS_u + EoS_p)) / (2 * Z + B * (EoS_u - EoS_p)), eps=eps_SL
        )
        return enth_ideal + enth_departure

    @staticmethod
    def enth_mol_phase_comp(blk, p, j):
        """
        Phase partial molar enthalpy
        """
        dlogphi_j_dT = _d_log_fug_coeff_dT_phase_comp(blk, p, j)

        enth_ideal_gas = get_method(blk, "enth_mol_ig_comp", j)(
            blk, cobj(blk, j), blk.temperature
        )

        enth_departure = -Cubic.gas_constant(blk) * blk.temperature**2 * dlogphi_j_dT

        return enth_ideal_gas + enth_departure

    @staticmethod
    def entr_mol_phase(blk, p):
        """
        Phase molar entropy
        """
        pobj = blk.params.get_phase(p)

        cname = pobj._cubic_type.name
        bm = getattr(blk, cname + "_bm")[p]
        B = getattr(blk, cname + "_B")[p]
        dam_dT = getattr(blk, cname + "_dam_dT")[p]
        Z = blk.compress_fact_phase[p]

        EoS_u = EoS_param[pobj._cubic_type]["u"]
        EoS_w = EoS_param[pobj._cubic_type]["w"]
        EoS_p = sqrt(EoS_u**2 - 4 * EoS_w)

        R = Cubic.gas_constant(blk)

        entr_ideal_gas = -R * safe_log(
            blk.pressure / blk.params.pressure_ref, eps=eps_SL
        )
        for j in blk.components_in_phase(p):
            entr_j = get_method(blk, "entr_mol_ig_comp", j)(
                blk, cobj(blk, j), blk.temperature
            )
            xj = blk.mole_frac_phase_comp[p, j]
            log_xj = blk.log_mole_frac_phase_comp[p, j]

            entr_ideal_gas += xj * (entr_j - R * log_xj)

        # See pg. 102 in Properties of Gases and Liquids
        # or pg. 208 of Sandler, 4th Ed.
        entr_departure = R * safe_log((Z - B), eps=eps_SL) + dam_dT / (
            bm * EoS_p
        ) * safe_log(
            (2 * Z + B * (EoS_u + EoS_p)) / (2 * Z + B * (EoS_u - EoS_p)), eps=eps_SL
        )

        return entr_ideal_gas + entr_departure

    @staticmethod
    def entr_mol_phase_comp(blk, p, j):
        """
        Phase partial molar entropy
        """
        logphi_j = _log_fug_coeff_phase_comp(blk, p, j)
        dlogphi_j_dT = _d_log_fug_coeff_dT_phase_comp(blk, p, j)

        R = Cubic.gas_constant(blk)

        entr_ideal_gas = get_method(blk, "entr_mol_ig_comp", j)(
            blk, cobj(blk, j), blk.temperature
        ) - R * (
            safe_log(blk.pressure / blk.params.pressure_ref, eps=eps_SL)
            + blk.log_mole_frac_phase_comp[p, j]
        )

        entr_departure = -R * logphi_j - R * blk.temperature * dlogphi_j_dT

        return entr_ideal_gas + entr_departure

    @staticmethod
    def fug_phase_comp(b, p, j):
        return b.mole_frac_phase_comp[p, j] * b.pressure * b.fug_coeff_phase_comp[p, j]

    @staticmethod
    def fug_phase_comp_eq(b, p, j, pp):
        return (
            b.mole_frac_phase_comp[p, j]
            * b.pressure
            * exp(_log_fug_coeff_phase_comp_eq(b, p, j, pp))
        )

    @staticmethod
    def log_fug_phase_comp_eq(b, p, j, pp):
        return (
            b.log_mole_frac_phase_comp[p, j]
            + log(b.pressure / b.params.pressure_ref)
            + _log_fug_coeff_phase_comp_eq(b, p, j, pp)
        )

    @staticmethod
    def fug_coeff_phase_comp(blk, p, j):
        pobj = blk.params.get_phase(p)
        ctype = pobj._cubic_type

        cname = pobj._cubic_type.name
        b = getattr(blk, cname + "_b")[j]
        bm = getattr(blk, cname + "_bm")[p]
        A = getattr(blk, cname + "_A")[p]
        B = getattr(blk, cname + "_B")[p]
        delta = getattr(blk, cname + "_delta")[p, j]
        Z = blk.compress_fact_phase[p]

        return exp(_log_fug_coeff_method(A, b, bm, B, delta, Z, ctype))

    @staticmethod
    def fug_coeff_phase_comp_eq(blk, p, j, pp):
        return exp(_log_fug_coeff_phase_comp_eq(blk, p, j, pp))

    @staticmethod
    def log_fug_phase_comp_Tbub(blk, p, j, pp):
        return _bubble_dew_log_fug_coeff_method(blk, p, j, pp, blk.temperature_bubble)

    @staticmethod
    def log_fug_phase_comp_Tdew(blk, p, j, pp):
        return _bubble_dew_log_fug_coeff_method(blk, p, j, pp, blk.temperature_dew)

    @staticmethod
    def log_fug_phase_comp_Pbub(blk, p, j, pp):
        return _bubble_dew_log_fug_coeff_method(blk, p, j, pp, blk.pressure_bubble)

    @staticmethod
    def log_fug_phase_comp_Pdew(blk, p, j, pp):
        return _bubble_dew_log_fug_coeff_method(blk, p, j, pp, blk.pressure_dew)

    @staticmethod
    def gibbs_mol_phase(b, p):
        return b.enth_mol_phase[p] - b.entr_mol_phase[p] * b.temperature

    @staticmethod
    def gibbs_mol_phase_comp(blk, p, j):
        # Calling the enthalpy and entropy directly adds a lot of overhead
        # because expressions involving the derivative of the fugacity coefficient
        # are generated. Those terms cancel mathematically, but I suspect Pyomo
        # leaves them in, causing trouble.
        R = Cubic.gas_constant(blk)
        T = blk.temperature
        logphi_j = _log_fug_coeff_phase_comp(blk, p, j)

        enth_ideal_gas = get_method(blk, "enth_mol_ig_comp", j)(blk, cobj(blk, j), T)

        entr_ideal_gas = get_method(blk, "entr_mol_ig_comp", j)(
            blk, cobj(blk, j), T
        ) - R * (
            safe_log(blk.pressure / blk.params.pressure_ref, eps=eps_SL)
            + blk.log_mole_frac_phase_comp[p, j]
        )
        gibbs_ideal_gas = enth_ideal_gas - T * entr_ideal_gas
        gibbs_departure = R * T * logphi_j

        return gibbs_ideal_gas + gibbs_departure

    @staticmethod
    def isentropic_speed_sound_phase(blk, p):
        # See Reference [2]
        return (
            sqrt(blk.heat_capacity_ratio_phase[p]) * blk.isothermal_speed_sound_phase[p]
        )

    @staticmethod
    def isothermal_speed_sound_phase(blk, p):
        pobj = blk.params.get_phase(p)
        cname = pobj._cubic_type.name
        am = getattr(blk, cname + "_am")[p]
        bm = getattr(blk, cname + "_bm")[p]
        V = 1 / blk.dens_mol_phase[p]
        mw = blk.mw
        rho = blk.dens_mass_phase[p]

        EoS_u = EoS_param[pobj._cubic_type]["u"]
        EoS_w = EoS_param[pobj._cubic_type]["w"]

        dPdV = -((Cubic.gas_constant(blk) * blk.temperature) / (V - bm) ** 2) + (
            am * (2 * V + EoS_u * bm) / (V**2 + EoS_u * bm * V + EoS_w * bm**2) ** 2
        )

        # see reference [2]
        return sqrt(-dPdV * mw / rho**2)

    @staticmethod
    def vol_mol_phase(b, p):
        return (
            Cubic.gas_constant(b)
            * b.temperature
            * b.compress_fact_phase[p]
            / b.pressure
        )

    @staticmethod
    def vol_mol_phase_comp(b, p, j):
        return (
            Cubic.gas_constant(b)
            * b.temperature
            / b.pressure
            * (b.compress_fact_phase[p] + _N_dZ_dNj(b, p, j))
        )

    @staticmethod
    def build_critical_properties(m, ref_phase):
        pobj = m.params.get_phase(ref_phase)
        cname = pobj.config.equation_of_state_options["type"].name
        ctype = pobj._cubic_type

        # Add components for calculation of mixture critical properties
        def func_a_crit(m, j):
            cobj = m.params.get_component(j)
            fw = getattr(m, cname + "_fw")
            return (
                EoS_param[ctype]["omegaA"]
                * (
                    (Cubic.gas_constant(m) * cobj.temperature_crit) ** 2
                    / cobj.pressure_crit
                )
                * (
                    (1 + fw[j] * (1 - sqrt(m.temperature_crit / cobj.temperature_crit)))
                    ** 2
                )
            )

        m.add_component(
            "a_crit",
            Expression(
                m.component_list, rule=func_a_crit, doc="Component a coefficient"
            ),
        )

        def rule_am_crit(m):
            try:
                rule = m.params.get_phase(ref_phase).config.equation_of_state_options[
                    "mixing_rule_a"
                ]
            except (KeyError, TypeError):
                rule = MixingRuleA.default

            a_crit = getattr(m, "a_crit")
            if rule == MixingRuleA.default:
                return rule_am_crit_default(m, cname, a_crit)
            else:
                raise ConfigurationError(
                    "{} Unrecognized option for Equation of State "
                    "mixing_rule_a: {}. Must be an instance of MixingRuleA "
                    "Enum.".format(m.name, rule)
                )

        m.add_component("am_crit", Expression(rule=rule_am_crit))

        def rule_bm_crit(m):
            try:
                rule = m.params.get_phase(ref_phase).config.equation_of_state_options[
                    "mixing_rule_b"
                ]
            except KeyError:
                rule = MixingRuleB.default

            b = getattr(m, cname + "_b")
            if rule == MixingRuleB.default:
                return rule_bm_crit_default(m, b)
            else:
                raise ConfigurationError(
                    "{} Unrecognized option for Equation of State "
                    "mixing_rule_b: {}. Must be an instance of MixingRuleB "
                    "Enum.".format(m.name, rule)
                )

        m.add_component("bm_crit", Expression(rule=rule_bm_crit))

        def rule_A_crit(m):
            am_crit = getattr(m, "am_crit")
            return EoS_param[ctype]["omegaA"] * (
                Cubic.gas_constant(m) * m.temperature_crit
            ) ** 2 == (am_crit * m.pressure_crit)

        m.add_component("A_crit", Constraint(rule=rule_A_crit))

        def rule_B_crit(m):
            bm_crit = getattr(m, "bm_crit")
            return EoS_param[ctype]["coeff_b"] * Cubic.gas_constant(
                m
            ) * m.temperature_crit == (bm_crit * m.pressure_crit)

        m.add_component("B_crit", Constraint(rule=rule_B_crit))

        Acrit = (
            m.am_crit
            * m.pressure_crit
            / (Cubic.gas_constant(m) * m.temperature_crit) ** 2
        )
        Bcrit = (
            m.bm_crit * m.pressure_crit / (Cubic.gas_constant(m) * m.temperature_crit)
        )

        expr_write = CubicThermoExpressions(m)
        if pobj.is_vapor_phase():
            Z_crit = expr_write.z_vap(eos=pobj._cubic_type, A=Acrit, B=Bcrit)
        elif pobj.is_liquid_phase():
            Z_crit = expr_write.z_liq(eos=pobj._cubic_type, A=Acrit, B=Bcrit)

        m.compress_fact_crit_eq = Constraint(rule=m.compress_fact_crit == Z_crit)

        m.dens_mol_crit_eq = Constraint(
            rule=m.pressure_crit
            == m.compress_fact_crit
            * Cubic.gas_constant(m)
            * m.temperature_crit
            * m.dens_mol_crit
        )

    @staticmethod
    def list_critical_property_constraint_names():
        return [
            "dens_mol_crit_eq",
            "compress_fact_crit_eq",
            "A_crit",
            "B_crit",
        ]


def _dZ_dT(blk, p):
    pobj = blk.params.get_phase(p)
    cname = pobj._cubic_type.name
    am = getattr(blk, cname + "_am")[p]
    A = getattr(blk, cname + "_A")[p]
    B = getattr(blk, cname + "_B")[p]
    dam_dT = getattr(blk, cname + "_dam_dT")[p]
    Z = blk.compress_fact_phase[p]
    T = blk.temperature

    EoS_u = EoS_param[pobj._cubic_type]["u"]
    EoS_w = EoS_param[pobj._cubic_type]["w"]

    dBdT = -B / T
    dAdT = (A / am) * dam_dT - (2 * A / T)

    K2 = (EoS_u - 1) * B - 1
    K3 = A - EoS_u * B - (EoS_u - EoS_w) * B**2

    dK2dT = (EoS_u - 1) * dBdT
    dK3dT = dAdT - EoS_u * dBdT - 2 * (EoS_u - EoS_w) * B * dBdT
    dK4dT = -(A * dBdT + B * dAdT + 2 * EoS_w * B * dBdT + 3 * EoS_w * B**2 * dBdT)

    return -(Z**2 * dK2dT + Z * dK3dT + dK4dT) / (3 * Z**2 + 2 * K2 * Z + K3)


def _N_dZ_dNj(blk, p, j):
    pobj = blk.params.get_phase(p)
    cname = pobj._cubic_type.name

    if not (
        pobj._mixing_rule_a == MixingRuleA.default
        and pobj._mixing_rule_b == MixingRuleB.default
    ):
        # Any user adding more mixing rules will need to explicitly add
        # support for these functions
        raise NotImplementedError(
            "Block {} called for a property "
            "that is not supported by this choice "
            "of mixing rules.".format(blk.name)
        )

    a = getattr(blk, cname + "_a")
    b = getattr(blk, cname + "_b")
    k = getattr(blk.params, cname + "_kappa")
    am = getattr(blk, cname + "_am")[p]
    bm = getattr(blk, cname + "_bm")[p]
    A = getattr(blk, cname + "_A")[p]
    B = getattr(blk, cname + "_B")[p]

    if pobj._mixing_rule_a == MixingRuleA.default:
        N_dam_dNj = 2 * (
            -am
            + sum(
                blk.mole_frac_phase_comp[p, i] * (1 - k[i, j]) * sqrt(a[i] * a[j])
                for i in blk.components_in_phase(p)
            )
        )

    if pobj._mixing_rule_b == MixingRuleB.default:
        N_dbm_dNj = b[j] - bm

    Z = blk.compress_fact_phase[p]
    R = Cubic.gas_constant(blk)
    T = blk.temperature
    P = blk.pressure

    EoS_u = EoS_param[pobj._cubic_type]["u"]
    EoS_w = EoS_param[pobj._cubic_type]["w"]

    N_dA_dNj = P / (R * T) ** 2 * N_dam_dNj
    N_dB_dNj = P / (R * T) * N_dbm_dNj

    K2 = (EoS_u - 1) * B - 1
    K3 = A - EoS_u * B - (EoS_u - EoS_w) * B**2

    N_dK2_dNj = (EoS_u - 1) * N_dB_dNj
    N_dK3_dNj = N_dA_dNj - EoS_u * N_dB_dNj - 2 * (EoS_u - EoS_w) * B * N_dB_dNj
    N_dK4_dNj = -(
        A * N_dB_dNj
        + B * N_dA_dNj
        + 2 * EoS_w * B * N_dB_dNj
        + 3 * EoS_w * B**2 * N_dB_dNj
    )

    return -(Z**2 * N_dK2_dNj + Z * N_dK3_dNj + N_dK4_dNj) / (
        3 * Z**2 + 2 * K2 * Z + K3
    )


def _log_fug_coeff_phase_comp_eq(blk, p, j, pp):
    pobj = blk.params.get_phase(p)

    cname = pobj._cubic_type.name
    b = getattr(blk, cname + "_b")
    bm = getattr(blk, cname + "_bm")
    Aeq = getattr(blk, "_" + cname + "_A_eq")
    Beq = getattr(blk, "_" + cname + "_B_eq")
    delta_eq = getattr(blk, "_" + cname + "_delta_eq")

    expr_write = CubicThermoExpressions(blk)
    if pobj.is_vapor_phase():

        def Zeq(p):
            return expr_write.z_vap(eos=pobj._cubic_type, A=Aeq[pp, p], B=Beq[pp, p])

    elif pobj.is_liquid_phase():

        def Zeq(p):
            return expr_write.z_liq(eos=pobj._cubic_type, A=Aeq[pp, p], B=Beq[pp, p])

    return _log_fug_coeff_method(
        Aeq[pp, p],
        b[j],
        bm[p],
        Beq[pp, p],
        delta_eq[pp, p, j],
        Zeq(p),
        pobj._cubic_type,
    )


def _log_fug_coeff_phase_comp(blk, p, j):
    pobj = blk.params.get_phase(p)

    if not (
        pobj._mixing_rule_a == MixingRuleA.default
        and pobj._mixing_rule_b == MixingRuleB.default
    ):
        # Any
        raise NotImplementedError(
            "Block {} called for a property "
            "that is not supported by this choice "
            "of mixing rules.".format(blk.name)
        )

    cname = pobj._cubic_type.name
    b = getattr(blk, cname + "_b")
    bm = getattr(blk, cname + "_bm")
    A = getattr(blk, cname + "_A")
    B = getattr(blk, cname + "_B")
    delta = getattr(blk, cname + "_delta")

    expr_write = CubicThermoExpressions(blk)
    if pobj.is_vapor_phase():

        def Z(p):
            return expr_write.z_vap(eos=pobj._cubic_type, A=A[p], B=B[p])

    elif pobj.is_liquid_phase():

        def Z(p):
            return expr_write.z_liq(eos=pobj._cubic_type, A=A[p], B=B[p])

    return _log_fug_coeff_method(
        A[p], b[j], bm[p], B[p], delta[p, j], Z(p), pobj._cubic_type
    )


def _log_fug_coeff_method(A, b, bm, B, delta, Z, cubic_type):
    u = EoS_param[cubic_type]["u"]
    w = EoS_param[cubic_type]["w"]
    p = sqrt(u**2 - 4 * w)

    return (
        b / bm * (Z - 1) * (B * p)
        - safe_log(Z - B, eps=eps_SL) * (B * p)
        + A
        * (b / bm - delta)
        * safe_log((2 * Z + B * (u + p)) / (2 * Z + B * (u - p)), eps=eps_SL)
    ) / (B * p)


def _d_log_fug_coeff_dT_phase_comp(blk, p, j):
    pobj = blk.params.get_phase(p)

    if not (
        pobj._mixing_rule_a == MixingRuleA.default
        and pobj._mixing_rule_b == MixingRuleB.default
    ):
        # Any
        raise NotImplementedError(
            "Block {} called for a property "
            "that is not supported by this choice "
            "of mixing rules.".format(blk.name)
        )

    cname = pobj._cubic_type.name
    am = getattr(blk, cname + "_am")[p]
    daij_dT = getattr(blk, cname + "_daij_dT")
    dam_dT = getattr(blk, cname + "_dam_dT")[p]
    b = getattr(blk, cname + "_b")[j]
    bm = getattr(blk, cname + "_bm")[p]
    A = getattr(blk, cname + "_A")[p]
    B = getattr(blk, cname + "_B")[p]
    delta = getattr(blk, cname + "_delta")[p, j]

    Z = blk.compress_fact_phase[p]
    dZ_dT = _dZ_dT(blk, p)
    T = blk.temperature

    u = EoS_param[pobj._cubic_type]["u"]
    w = EoS_param[pobj._cubic_type]["w"]
    EoS_p = sqrt(u**2 - 4 * w)

    expr = (
        A
        / (EoS_p * B)
        * (
            b / bm * (1 / am * dam_dT - 1 / T)
            + delta / T
            - 2
            / am
            * sum(
                blk.mole_frac_phase_comp[p, i] * daij_dT[i, j]
                for i in blk.component_list
            )
        )
    )

    return (
        b / bm * dZ_dT
        - (dZ_dT + B / T) / (Z - B)
        - A * (b / bm - delta) * (Z / T + dZ_dT) / (Z**2 + u * B * Z + w * B**2)
        + log((2 * Z + B * (u + EoS_p)) / (2 * Z + B * (u - EoS_p))) * expr
    )


def _bubble_dew_log_fug_coeff_method(blk, p, j, pp, pt_var):
    pobj = blk.params.get_phase(p)
    ctype = pobj._cubic_type
    cname = pobj.config.equation_of_state_options["type"].name

    # Ditch the m.fs.unit.control_volume...
    short_name = pt_var.name.split(".")[-1]
    # import pdb; pdb.set_trace()
    if short_name.startswith("temperature"):
        abbrv = "t"
        T = pt_var[pp]
        P = blk.pressure
    elif short_name.startswith("pressure"):
        abbrv = "p"
        P = pt_var[pp]
        T = blk.temperature
    else:
        raise BurntToast(
            "Users shouldn't be calling this function. "
            "If you're a dev, you know what you did."
        )

    if short_name.endswith("bubble"):
        abbrv += "bub"
        if pobj.is_liquid_phase():
            x = blk.mole_frac_comp
            log_mole_frac = blk.log_mole_frac_comp
            xidx = ()
        elif pobj.is_vapor_phase():
            x = getattr(blk, "_mole_frac_" + abbrv)
            log_mole_frac = getattr(blk, "log_mole_frac_" + abbrv)
            xidx = pp
        else:
            raise BurntToast(
                "Users shouldn't be calling this function. "
                "If you're a dev, you know what you did."
            )

    elif short_name.endswith("dew"):
        abbrv += "dew"
        if pobj.is_vapor_phase():
            x = blk.mole_frac_comp
            log_mole_frac = blk.log_mole_frac_comp
            xidx = ()
        elif pobj.is_liquid_phase():
            x = getattr(blk, "_mole_frac_" + abbrv)
            log_mole_frac = getattr(blk, "log_mole_frac_" + abbrv)
            xidx = pp
        else:
            raise BurntToast(
                "Users shouldn't be calling this function. "
                "If you're a dev, you know what you did."
            )
    else:
        raise BurntToast(
            "Users shouldn't be calling this function. "
            "If you're a dev, you know what you did."
        )

    def a(k):
        cobj = blk.params.get_component(k)
        fw = getattr(blk, cname + "_fw")[k]
        ac = getattr(blk, cname + "_a_crit")[k]
        func_alpha = getattr(blk.params, cname + "_func_alpha")

        return ac * func_alpha(T, fw, cobj)

    kappa = getattr(blk.params, cname + "_kappa")
    am = sum(
        sum(
            x[xidx, i] * x[xidx, j] * sqrt(a(i) * a(j)) * (1 - kappa[i, j])
            for j in blk.component_list
        )
        for i in blk.component_list
    )

    b = getattr(blk, cname + "_b")
    bm = sum(x[xidx, i] * b[i] for i in blk.component_list)
    R = Cubic.gas_constant(blk)

    A = am * P / (R * T) ** 2
    B = bm * P / (R * T)

    delta = (
        2
        * sqrt(a(j))
        / am
        * sum(x[xidx, i] * sqrt(a(i)) * (1 - kappa[j, i]) for i in blk.component_list)
    )

    expr_write = CubicThermoExpressions(blk)
    if pobj.is_vapor_phase():
        Z = expr_write.z_vap(eos=pobj._cubic_type, A=A, B=B)
    elif pobj.is_liquid_phase():
        Z = expr_write.z_liq(eos=pobj._cubic_type, A=A, B=B)

    return (
        _log_fug_coeff_method(A, b[j], bm, B, delta, Z, ctype)
        + log_mole_frac[xidx, j]
        + log(P / blk.params.pressure_ref)
    )


# -----------------------------------------------------------------------------
# Default rules for cubic expressions
def calculate_equilibrium_cubic_coefficients(b, cubic_name, cubic_type, p1, p2, p3):
    """
    Calculates the coefficients b, c, and d of the cubic 0 = z**3 + b * z**2 + c * z + d
    at the equilibrium conditions

    Args:
        b: StateBlock of interest
        cubic_name: Name of Cubic EoS
        cubic_type: Type of Cubic EoS
        p1: Phase 1
        p2: Phase 2
        p3; Phase 3

    Returns:
        expressions for b, c, d
    """

    A_eq = getattr(b, "_" + cubic_name + "_A_eq")[p1, p2, p3]
    B_eq = getattr(b, "_" + cubic_name + "_B_eq")[p1, p2, p3]
    EoS_u = EoS_param[cubic_type]["u"]
    EoS_w = EoS_param[cubic_type]["w"]

    b = -(1 + B_eq - EoS_u * B_eq)
    c = A_eq - EoS_u * B_eq - EoS_u * B_eq**2 + EoS_w * B_eq**2
    d = -(A_eq * B_eq + EoS_w * B_eq**2 + EoS_w * B_eq**3)

    return b, c, d


def func_fw_PR(cobj):
    """
    f(omega) function for Peng-Robinson EoS.

    Args:
        cobj: Component object

    Returns:
        expression for fw

    """
    return 0.37464 + 1.54226 * cobj.omega - 0.26992 * cobj.omega**2


def func_fw_SRK(cobj):
    """
    f(omega) function for SRK EoS.

    Args:
        cobj: Component object

    Returns:
        expression for fw

    """
    return 0.48 + 1.574 * cobj.omega - 0.176 * cobj.omega**2


def func_alpha_soave(T, fw, cobj):
    """
    Soave alpha function.

    Args:
        fw: expression for fw
        cobj: Component object

    Returns:
        expression for alpha

    """
    Tc = cobj.temperature_crit
    Tr = T / Tc
    return (1 + fw * (1 - sqrt(Tr))) ** 2


def func_dalpha_dT_soave(T, fw, cobj):
    """
    Function to get first partial derivative w.r.t. temperature of Soave alpha function.

    Args:
        fw: expression for fw
        cobj: Component object

    Returns:
        expression for first derivative of alpha

    """
    Tc = cobj.temperature_crit
    Tr = T / Tc
    return 1 / Tc * (-fw / sqrt(Tr)) * (1 + fw * (1 - sqrt(Tr)))


def func_d2alpha_dT2_soave(T, fw, cobj):
    """
    Function to get 2nd partial derivative w.r.t. temperature of Soave alpha function.

    Args:
        fw: expression for fw
        cobj: Component object

    Returns:
        expression for 2nd derivative of alpha

    """
    Tc = cobj.temperature_crit
    Tr = T / Tc
    return 1 / Tc**2 * ((fw**2 + fw) / (2 * Tr * sqrt(Tr)))


# -----------------------------------------------------------------------------
# Mixing rules
def rule_am_default(m, cname, a, p, pp=()):
    """
    Standard Van der Waals one-fluid mixing rule for a term
    """
    k = getattr(m.params, cname + "_kappa")
    return sum(
        sum(
            m.mole_frac_phase_comp[p, i]
            * m.mole_frac_phase_comp[p, j]
            * sqrt(a[pp, i] * a[pp, j])
            * (1 - k[i, j])
            for j in m.components_in_phase(p)
        )
        for i in m.components_in_phase(p)
    )


def rule_am_crit_default(m, cname, a_crit):
    """
    Standard Van der Waals one-fluid mixing rule for a term evaluated at critical point
    """
    k = getattr(m.params, cname + "_kappa")
    return sum(
        sum(
            m.mole_frac_comp[i]
            * m.mole_frac_comp[j]
            * sqrt(a_crit[i] * a_crit[j])
            * (1 - k[i, j])
            for j in m.component_list
        )
        for i in m.component_list
    )


def rule_bm_default(m, b, p):
    """
    Standard Van der Waals one-fluid mixing rule for b term
    """
    return sum(m.mole_frac_phase_comp[p, i] * b[i] for i in m.components_in_phase(p))


def rule_bm_crit_default(m, b):
    """
    Standard Van der Waals one-fluid mixing rule for b term evaluated at critical point
    """
    return sum(m.mole_frac_comp[i] * b[i] for i in m.component_list)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Methods for defining reaction rates
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

from idaes.models.properties.modular_properties.base.utility import (
    get_concentration_term,
)


# -----------------------------------------------------------------------------
class power_law_rate:
    """Methods for power-law rate expressions."""

    @staticmethod
    def build_parameters(rblock, config):
        pass

    @staticmethod
    def return_expression(b, rblock, r_idx, T):
        e = None
        # Get reaction orders and construct power law expression
        for p, j in b.phase_component_set:
            o = rblock.reaction_order[p, j]

            if e is None and o.value != 0:
                e = get_concentration_term(b, r_idx)[p, j] ** o
            elif e is not None and o.value != 0:
                e = e * get_concentration_term(b, r_idx)[p, j] ** o

        return b.k_rxn[r_idx] * e



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Predefined expression for Helmholtz EoS functions
"""

__author__ = "John Eslick"

from idaes.core.util.math import smooth_max


def surface_tension_type01(model, parameters):
    """Type01 expression for the surface tension

    Args:
        model (Block): Pyomo model
        parameters (dict): Main parameters dictionary

    Returns:
        dict: Expressions for the surface tension
    """
    s = parameters["transport"]["surface_tension"]["s"]
    n = parameters["transport"]["surface_tension"]["n"]
    tc = parameters["transport"]["surface_tension"]["Tc"]
    eps = 1e-4
    # Use smooth max to avoid evaluation errors on negative numbers
    # this means the surface tension will go to zero at and above the
    # critical temperature, rather than cause error over critical.
    omTr = smooth_max(1 - model.T_star / model.tau / tc, 0, eps=eps)
    return sum(s[i] * omTr ** n[i] for i in s)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Standard IDAES Gibbs reactor model.
"""
# Import Pyomo libraries
from pyomo.environ import Constraint, Param, Reals, Reference, Set, units, value, Var
from pyomo.common.config import ConfigBlock, ConfigValue, In, ListOf, Bool

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.scaling import CustomScalerBase
from idaes.core.util.constants import Constants

__author__ = "Jinliang Ma, Andrew Lee"


class GibbsReactorScaler(CustomScalerBase):
    """
    Scaler for Gibbs Reactor units.

    Due to the nature of Gibbs Reactors, scaling is highly dependent on the outlet
    concentrations which cannot be predicted a priori, thus we rely on users to
    provide the best initial guesses they can for the outlet concentrations.

    """

    UNIT_SCALING_FACTORS = {
        # "QuantityName: (reference units, scaling factor)
        "Delta Pressure": (units.Pa, 1e-3),
        "Heat": (units.J / units.s, 1e-6),
    }

    def variable_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        """
        Variable scaling routine for Gibbs reactors.

        Due to the nature of Gibbs Reactors, scaling is highly dependent on the outlet
        concentrations which cannot be predicted a priori, thus we rely on users to
        provide the best initial guesses they can for the outlet concentrations.

        Args:
            model: instance of GibbsReactor to be scaled
            overwrite: whether to overwrite existing scaling factors
            submodel_scalers: dict of Scalers to use for sub-models, keyed by submodel local name

        Returns:
            None
        """
        if submodel_scalers is None:
            submodel_scalers = {}

        # Step 1: Property scaling

        # Step 1a: propagate any existing scaling from inlet to outlet
        # This is likely a very poor approximation for concentrations, but we expect users
        # to provide better scaling factors manually (as there is no way for us to know).
        # This will do a first pass fill in however if the user does not provide any information.
        self.propagate_state_scaling(
            target_state=model.control_volume.properties_out,
            source_state=model.control_volume.properties_in,
            overwrite=overwrite,
        )

        # Step 1b: Call Scalers for state blocks
        # Inlet properties
        self.call_submodel_scaler_method(
            submodel=model.control_volume.properties_in,
            submodel_scalers=submodel_scalers,
            method="variable_scaling_routine",
            overwrite=overwrite,
        )
        # Outlet properties
        self.call_submodel_scaler_method(
            submodel=model.control_volume.properties_out,
            submodel_scalers=submodel_scalers,
            method="variable_scaling_routine",
            overwrite=overwrite,
        )

        # Step 2: Scaling Gibbs reactor variables
        # Control volume variables - support only heat and deltaP
        if hasattr(model.control_volume, "heat"):
            for v in model.control_volume.heat.values():
                self.scale_variable_by_units(v, overwrite=overwrite)
        if hasattr(model.control_volume, "deltaP"):
            for v in model.control_volume.deltaP.values():
                self.scale_variable_by_units(v, overwrite=overwrite)

        # Lagrangian multipliers
        # Best guess scaling for these is R*T, need to convert units
        p_units = (
            model.control_volume.config.property_package.get_metadata().get_derived_units
        )
        for (t, _), v in model.lagrange_mult.items():
            tsf = self.get_scaling_factor(
                model.control_volume.properties_out[t].temperature
            )
            if tsf is not None:
                nominal_t = 1 / tsf
            else:
                nominal_t = 500
            lsf = value(
                1
                / units.convert(
                    Constants.gas_constant * nominal_t * p_units("temperature"),
                    to_units=p_units("energy_mole"),
                )
            )
            self.set_variable_scaling_factor(v, lsf, overwrite=overwrite)

    def constraint_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        """
        Routine to apply scaling factors to constraints in model.

        Constraints will be scaled based on nominal Jacobian norms, and thus will
        be heavily dependent on variable scaling.

        Args:
            model: model to be scaled
            overwrite: whether to overwrite existing scaling factors
            submodel_scalers: dict of Scalers to use for sub-models, keyed by submodel local name

        Returns:
            None
        """
        # Step 1: Call Scalers for state blocks
        # Inlet properties
        self.call_submodel_scaler_method(
            submodel=model.control_volume.properties_in,
            submodel_scalers=submodel_scalers,
            method="constraint_scaling_routine",
            overwrite=overwrite,
        )
        # Outlet properties
        self.call_submodel_scaler_method(
            submodel=model.control_volume.properties_out,
            submodel_scalers=submodel_scalers,
            method="constraint_scaling_routine",
            overwrite=overwrite,
        )

        # Step 2: Scale all the control volume constraints
        for cd in model.control_volume.component_data_objects(
            ctype=Constraint, descend_into=False
        ):
            self.scale_constraint_by_nominal_value(
                cd, scheme="inverse_sum", overwrite=overwrite
            )

        # Step 3: Scale local constraints
        # Scale Gibbs minimization constraints
        for cd in model.gibbs_minimization.values():
            self.scale_constraint_by_nominal_value(
                cd, scheme="inverse_sum", overwrite=overwrite
            )

        # Scale inert species balance if they are present
        if hasattr(model, "inert_species_balance"):
            for cd in model.inert_species_balance.values():
                self.scale_constraint_by_nominal_value(
                    cd, scheme="inverse_sum", overwrite=overwrite
                )


@declare_process_block_class("GibbsReactor")
class GibbsReactorData(UnitModelBlockData):
    """
    Standard Gibbs Reactor Unit Model Class

    This model assume all possible reactions reach equilibrium such that the
    system partial molar Gibbs free energy is minimized.
    Since some species mole flow rate might be very small,
    the natural log of the species molar flow rate is used.
    Instead of specifying the system Gibbs free energy as an objective
    function, the equations for zero partial derivatives of the grand function
    with Lagrangian multiple terms with respect to product species mole flow
    rates and the multiples are specified as constraints.
    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Gibbs reactors do not support dynamic models, thus this must be
False.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag",
            doc="""Gibbs reactors do not have defined volume, thus this must be
False.""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.useDefault.
**Valid values:** {
**EnergyBalanceType.useDefault - refer to property package for default
balance type
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_heat_transfer",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Heat transfer term construction flag",
            doc="""Indicates whether terms for heat transfer should be constructed,
**default** - False.
**Valid values:** {
**True** - include heat transfer terms,
**False** - exclude heat transfer terms.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "inert_species",
        ConfigValue(
            default=[],
            domain=ListOf(str),
            description="List of inert species",
            doc="List of species which do not take part in reactions.",
        ),
    )

    default_scaler = GibbsReactorScaler

    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(GibbsReactorData, self).build()

        # Build Control Volume
        self.control_volume = ControlVolume0DBlock(
            dynamic=self.config.dynamic,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.control_volume.add_state_blocks(has_phase_equilibrium=False)

        # Validate list of inert species
        for i in self.config.inert_species:
            if i not in self.control_volume.properties_in.component_list:
                raise ConfigurationError(
                    "{} invalid component in inert_species argument. {} is "
                    "not in the property package component list.".format(self.name, i)
                )

        self.control_volume.add_total_element_balances()

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=self.config.has_heat_transfer,
        )

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change,
        )

        # Add Ports
        self.add_inlet_port()
        self.add_outlet_port()

        # Add performance equations
        # Add Lagrangian multiplier variables
        # Only need multipliers for species which participate in reactions
        l_set = []
        for e in self.control_volume.config.property_package.element_list:
            skip = True
            for j in self.control_volume.properties_in.component_list:
                if j in self.config.inert_species:
                    continue
                elif self.control_volume.properties_out.params.element_comp[j][e] != 0:
                    skip = False

            if not skip:
                l_set.append(e)
        self.lagrange_set = Set(initialize=l_set)

        e_units = self.control_volume.config.property_package.get_metadata().get_derived_units(
            "energy_mole"
        )
        self.lagrange_mult = Var(
            self.flowsheet().time,
            self.lagrange_set,
            domain=Reals,
            initialize=100,
            doc="Lagrangian multipliers",
            units=e_units,
        )

        # TODO : Remove this once scaling is properly implemented
        self.gibbs_scaling = Param(default=1, mutable=True)

        # Use Lagrangian multiple method to derive equations for Out_Fi
        # Use RT*lagrange as the Lagrangian multiple such that lagrange is in
        # a similar order of magnitude as log(Yi)

        @self.Constraint(
            self.flowsheet().time,
            self.control_volume.properties_in.phase_component_set,
            doc="Gibbs energy minimisation constraint",
        )
        def gibbs_minimization(b, t, p, j):
            # Use natural log of species mole flow to avoid Pyomo solver
            # warnings of reaching infeasible point
            if j in self.config.inert_species:
                return Constraint.Skip
            return 0 == b.gibbs_scaling * (
                b.control_volume.properties_out[t].gibbs_mol_phase_comp[p, j]
                + sum(
                    b.lagrange_mult[t, e]
                    * b.control_volume.properties_out[t].params.element_comp[j][e]
                    for e in self.lagrange_set
                )
            )

        if len(self.config.inert_species) > 0:

            @self.Constraint(
                self.flowsheet().time,
                self.control_volume.properties_in.phase_list,
                self.config.inert_species,
                doc="Inert species balances",
            )
            def inert_species_balance(b, t, p, j):
                # Add species balances for inert components
                cv = b.control_volume
                e_comp = cv.properties_out[t].config.parameters.element_comp

                # Check for linear dependence with element balances
                # If an inert species is the only source of element e,
                # the inert species balance would be linearly dependent on the
                # element balance for e.
                dependent = True

                if len(self.control_volume.properties_in.phase_list) > 1:
                    # Multiple phases avoid linear dependency
                    dependent = False
                else:
                    for e in self.control_volume.config.property_package.element_list:
                        if e_comp[j][e] == 0:
                            # Element e not in component j, no effect
                            continue
                        else:
                            for i in self.control_volume.properties_in.component_list:
                                if i == j:
                                    continue
                                else:
                                    # If comp j shares element e with comp i
                                    # cannot be linearly dependent
                                    if e_comp[i][e] != 0:
                                        dependent = False

                if (
                    not dependent
                    and (p, j) in self.control_volume.properties_in.phase_component_set
                ):
                    return 0 == (
                        cv.properties_in[t].get_material_flow_terms(p, j)
                        - cv.properties_out[t].get_material_flow_terms(p, j)
                    )
                else:
                    return Constraint.Skip

        # Set references to balance terms at unit level
        if (
            self.config.has_heat_transfer is True
            and self.config.energy_balance_type != EnergyBalanceType.none
        ):
            self.heat_duty = Reference(self.control_volume.heat[:])
        if (
            self.config.has_pressure_change is True
            and self.config.momentum_balance_type != MomentumBalanceType.none
        ):
            self.deltaP = Reference(self.control_volume.deltaP[:])

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        if hasattr(self, "heat_duty"):
            var_dict["Heat Duty"] = self.heat_duty[time_point]
        if hasattr(self, "deltaP"):
            var_dict["Pressure Change"] = self.deltaP[time_point]

        return {"vars": var_dict}
